# type: ignore
import inspect
import re
import typing

import pytest
from langchain.chat_models.base import BaseChatModel
from langchain.llms import BaseLLM
from pydantic import BaseModel, Field, create_model
from pydantic.generics import GenericModel

from llm_strategy import llm_function
from llm_strategy.chat_chain import ChatChain
from llm_strategy.llm_function import (
    LLMBoundSignature,
    LLMStructuredPrompt,
    Output,
    get_concise_type_repr,
    is_not_implemented,
)
from llm_strategy.testing.fake_llm import FakeLLM


def not_implemented_function():
    raise NotImplementedError


def test_is_not_implemented_function():
    assert is_not_implemented(not_implemented_function)
    assert not is_not_implemented(lambda: 1)


def test_get_concise_type_repr():
    assert get_concise_type_repr(int) == "int"
    assert get_concise_type_repr(typing.List[int]) == "list[int]"
    assert get_concise_type_repr(list[int]) == "list[int]"
    assert get_concise_type_repr(typing.List[typing.List[int]]) == "list[list[int]]"
    assert get_concise_type_repr(list[list[int]]) == "list[list[int]]"
    assert get_concise_type_repr(typing.Dict[str, int]) == "dict[str, int]"
    assert get_concise_type_repr(dict[str, int]) == "dict[str, int]"
    assert get_concise_type_repr(typing.Dict[str, typing.List[int]]) == "dict[str, list[int]]"
    assert get_concise_type_repr(dict[str, list[int]]) == "dict[str, list[int]]"
    assert get_concise_type_repr(typing.Dict[str, typing.Dict[str, int]]) == "dict[str, dict[str, int]]"
    assert get_concise_type_repr(typing.Set[int]) == "set[int]"
    assert get_concise_type_repr(set[int]) == "set[int]"
    assert get_concise_type_repr(typing.Set[typing.Set[int]]) == "set[set[int]]"
    assert get_concise_type_repr(set[set[int]]) == "set[set[int]]"
    assert get_concise_type_repr(typing.Set[typing.List[int]]) == "set[list[int]]"

    class A:
        pass

    assert get_concise_type_repr(A) == "A"
    assert get_concise_type_repr(typing.List[A]) == "list[A]"

    # generic class
    T = typing.TypeVar("T")

    class B(GenericModel, typing.Generic[T]):
        pass

    assert get_concise_type_repr(B) == "B[T]"
    assert get_concise_type_repr(typing.List[B]) == "list[B[T]]"
    assert get_concise_type_repr(typing.List[B[int]]) == "list[B[int]]"
    assert get_concise_type_repr(list[B[int]]) == "list[B[int]]"


def test_llm_function_first_param():
    def f(llm: BaseLLM, a: str = "", b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    assert LLMBoundSignature.from_call(f, (), {}).input_type.schema() == {
        "properties": {
            "a": {"default": "", "title": "A", "type": "string"},
            "b": {"default": 1, "title": "B", "type": "integer"},
        },
        "title": "FInputs",
        "type": "object",
    }

    def g(chat_model: BaseChatModel, a: str = "", b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    assert LLMBoundSignature.from_call(g, (), {}).input_type.schema() == {
        "properties": {
            "a": {"default": "", "title": "A", "type": "string"},
            "b": {"default": 1, "title": "B", "type": "integer"},
        },
        "title": "GInputs",
        "type": "object",
    }

    def h(chat_chain: ChatChain, a: str = "", b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    assert LLMBoundSignature.from_call(h, (), {}).input_type.schema() == {
        "properties": {
            "a": {"default": "", "title": "A", "type": "string"},
            "b": {"default": 1, "title": "B", "type": "integer"},
        },
        "title": "HInputs",
        "type": "object",
    }

    # with a wrong type
    with pytest.raises(ValueError):

        def i(x: int, a: str = "", b: int = 1) -> str:
            """Test docstring."""
            raise NotImplementedError

        LLMBoundSignature.from_call(i, (), {})


def test_llm_bound_signature_from_call():
    def f(llm: BaseLLM, a: str, b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(f, ("",), {})
    assert llm_bound_signature.docstring == "Test docstring."
    assert llm_bound_signature.signature == inspect.signature(f)
    assert llm_bound_signature.input_type.schema() == create_model("FInputs", a=(str, ...), b=(int, 1)).schema()
    assert llm_bound_signature.output_type.schema() == create_model("Output[str]", return_value=(str, ...)).schema()


def test_llm_function_from_call_first_param():
    def f(llm: BaseLLM, a: str, b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    assert LLMBoundSignature.from_call(f, ("",), dict(b=1)).input_type.schema() == {
        "properties": {
            "a": {"title": "A", "type": "string"},
            "b": {"title": "B", "type": "integer", "default": 1},
        },
        "required": ["a"],
        "title": "FInputs",
        "type": "object",
    }

    def g(chat_model: BaseChatModel, a: str, b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    assert LLMBoundSignature.from_call(g, ("",), {}).input_type.schema() == {
        "properties": {
            "a": {"title": "A", "type": "string"},
            "b": {"title": "B", "type": "integer", "default": 1},
        },
        "required": ["a"],
        "title": "GInputs",
        "type": "object",
    }

    def h(chat_chain: ChatChain, a: str, b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    assert LLMBoundSignature.from_call(h, ("",), {}).input_type.schema() == {
        "properties": {
            "a": {"title": "A", "type": "string"},
            "b": {"title": "B", "type": "integer", "default": 1},
        },
        "required": ["a"],
        "title": "HInputs",
        "type": "object",
    }

    # with a wrong type
    with pytest.raises(ValueError):

        def i(x: int, a: str, b: int = 1) -> str:
            """Test docstring."""
            raise NotImplementedError

        LLMBoundSignature.from_call(i, ("",), {})


def test_llm_bound_signature_from_call_with_field():
    # Use Pydantic's Field to specify a default value.
    def f(llm: BaseLLM, a: str, b=Field(3)) -> str:  # noqa: B008
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(f, ("",), {})

    assert llm_bound_signature.input_type.schema() == create_model("FInputs", a=(str, ...), b=(int, 3)).schema()


def test_llm_bound_signature_from_call_with_missing_default():
    # Use Pydantic's Field to specify a description.
    def f(llm: BaseLLM, a: str, b: int = Field(..., description="test")) -> str:  # noqa: B008
        """Test docstring."""
        raise NotImplementedError

    with pytest.raises(TypeError, match=re.escape("missing a required argument: 'b'")):
        LLMBoundSignature.from_call(f, ("",), {})


def test_llm_bound_signature_from_call_with_field_description_no_default():
    # Use Pydantic's Field to specify a description.
    def f(llm: BaseLLM, a: str, b: int = Field(..., description="test")) -> str:  # noqa: B008
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(f, ("",), dict(b=1))

    assert llm_bound_signature.input_type.schema() == {
        "properties": {
            "a": {"title": "A", "type": "string"},
            "b": {"description": "test", "title": "B", "type": "integer"},
        },
        "required": ["a", "b"],
        "title": "FInputs",
        "type": "object",
    }


def test_llm_bound_signature_from_call_no_docstring():
    def f(llm: BaseLLM, a: str, b: int = 1) -> str:
        raise NotImplementedError

    with pytest.raises(ValueError):
        LLMBoundSignature.from_call(f, ("",), {})


def test_llm_bound_signature_from_call_no_return_type():
    def f(llm: BaseLLM, a: str, b: int = 1):
        """Test docstring."""
        raise NotImplementedError

    with pytest.raises(ValueError):
        LLMBoundSignature.from_call(f, ("",), {})


def test_llm_bound_signature_from_generic_call_to_none():
    T = typing.TypeVar("T")

    def f(llm: BaseLLM, a: T) -> T:
        """Test docstring."""
        raise NotImplementedError

    with pytest.raises(ValueError):
        LLMBoundSignature.from_call(f, (None,), {})


def test_llm_bound_signature_from_call_no_parameter_annotation_but_default():
    def f(llm: BaseLLM, a=1, b: int = 1) -> str:
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(f, (), {})
    assert llm_bound_signature.input_type.schema() == create_model("FInputs", a=(int, 1), b=(int, 1)).schema()


def test_llm_bound_signature_from_call_generic_input_outputs() -> None:
    T = typing.TypeVar("T")
    V = typing.TypeVar("V")

    class GenericType(GenericModel, typing.Generic[T]):
        value: T

    class GenericType2(GenericModel, typing.Generic[T, V]):
        value: T
        value2: V

    def f(llm: BaseLLM, a: GenericType[T], b: GenericType[V]) -> GenericType2[T, V]:
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(
        f, (), dict(a=GenericType[int](value=0), b=GenericType[str](value=""))
    )
    assert llm_bound_signature.output_type.schema() == Output[GenericType2[int, str]].schema()


def test_llm_bound_signature_from_call_specified_generic_parameters() -> None:
    def f(llm: BaseLLM, a: list[str], b: dict[str, list[str]]) -> dict[str, str]:
        """Test docstring."""
        raise NotImplementedError

    class FInputs(BaseModel):
        a: list[str]
        b: dict[str, list[str]]

    llm_bound_signature = LLMBoundSignature.from_call(f, (), dict(a=["a"], b=dict(t=["b"])))
    assert llm_bound_signature.input_type.schema() == FInputs.schema()
    assert llm_bound_signature.output_type.schema() == Output[dict[str, str]].schema()


def test_llm_bound_signature_from_call_generic_collection() -> None:
    T = typing.TypeVar("T")

    def f(llm: BaseLLM, a: list[T], b: T) -> dict[T, T]:
        """Test docstring."""
        raise NotImplementedError

    class FInputs(GenericModel, typing.Generic[T]):
        a: list[T]
        b: T

    llm_bound_signature = LLMBoundSignature.from_call(f, (), dict(a=["a"], b="hello"))
    assert llm_bound_signature.input_type.schema() == FInputs[str].schema()
    assert llm_bound_signature.output_type.schema() == Output[dict[str, str]].schema()


def test_llm_bound_signature_from_call_generic_collection_2() -> None:
    # TODO: I do not like the result but I don't want to implement a full generic type resolution system.
    T = typing.TypeVar("T")

    def f(llm: BaseLLM, a: list[T]) -> dict[T, T]:
        """Test docstring."""
        raise NotImplementedError

    class FInputs(GenericModel, typing.Generic[T]):
        a: list[T]

    llm_bound_signature = LLMBoundSignature.from_call(f, (), dict(a=["a"]))
    assert llm_bound_signature.input_type.schema() == FInputs.schema()
    assert llm_bound_signature.output_type.schema() == Output[dict[T, T]].schema()


def test_llm_bound_signature_from_call_generic_function() -> None:
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    def f(llm: BaseLLM, a: T, b: S) -> dict[T, S]:
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(f, (), dict(a=0, b=""))
    assert llm_bound_signature.output_type.schema() == Output[dict[int, str]].schema()


def test_llm_bound_signature_from_call_generic_function_with_primitive_types() -> None:
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    def f(llm: BaseLLM, a: T, b: S, c: int) -> dict[T, S]:
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(f, (), dict(a=0, b="", c=10))
    assert llm_bound_signature.output_type.schema() == Output[dict[int, str]].schema()


def test_llm_bound_signature_from_call_generic_input_outputs_full_remap() -> None:
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    U = typing.TypeVar("U")
    V = typing.TypeVar("V")

    class GenericType(GenericModel, typing.Generic[T]):
        value: T

    class GenericType2(GenericModel, typing.Generic[T, S]):
        value: T
        value2: S

    def f(llm: BaseLLM, a: GenericType[U], b: GenericType[V], c: GenericType[V]) -> GenericType2[U, V]:
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(
        f, (), dict(a=GenericType[int](value=0), b=GenericType[str](value=""), c=GenericType[str](value=""))
    )
    assert llm_bound_signature.output_type.schema() == Output[GenericType2[int, str]].schema()


def test_llm_bound_signature_from_call_generic_input_outputs_multiple_remap() -> None:
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    U = typing.TypeVar("U")
    V = typing.TypeVar("V")

    class GenericType(GenericModel, typing.Generic[T, S]):
        value: T | None = None
        value2: S | None = None

    def f(
        llm: BaseLLM, a: GenericType[U, int], b: GenericType[int, V], c: GenericType[V, V], d: GenericType[U, U]
    ) -> GenericType[U, V]:
        """Test docstring."""
        raise NotImplementedError

    llm_bound_signature = LLMBoundSignature.from_call(
        f,
        (),
        dict(
            a=GenericType[int, int](), b=GenericType[int, str](), c=GenericType[str, str](), d=GenericType[int, int]()
        ),
    )
    assert llm_bound_signature.output_type.schema() == Output[GenericType[int, str]].schema()


def test_llm_bound_signature_from_call_generic_input_outputs_full_remap_failed() -> None:
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    U = typing.TypeVar("U")
    V = typing.TypeVar("V")

    class GenericType(GenericModel, typing.Generic[T]):
        value: T

    class GenericType2(GenericModel, typing.Generic[T, S]):
        value: T
        value2: S

    def f(llm: BaseLLM, a: GenericType[U], b: GenericType[V], c: GenericType[V]) -> GenericType2[U, V]:
        """Test docstring."""
        raise NotImplementedError

    with pytest.raises(
        ValueError,
        match=re.escape("Cannot resolve generic type ~V, conflicting resolution: <class 'str'> and <class 'float'>."),
    ):
        LLMBoundSignature.from_call(
            f,
            (),
            dict(a=GenericType[int](value=0), b=GenericType[str](value=""), c=GenericType[float](value=0.0)),
        )


def test_get_generic_type_map() -> None:
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    U = typing.TypeVar("U")
    V = typing.TypeVar("V")

    X = typing.TypeVar("X")

    class GenericType(GenericModel, typing.Generic[T, S]):
        value: T
        value2: S

    assert LLMStructuredPrompt.get_generic_type_map(GenericType) == {T: T, S: S}
    assert LLMStructuredPrompt.get_generic_type_map(GenericType[S, T]) == {T: S, S: T}
    assert LLMStructuredPrompt.get_generic_type_map(GenericType[S, T][U, V]) == {T: U, S: V}  # type: ignore
    assert LLMStructuredPrompt.get_generic_type_map(GenericType[S, T][U, V][X, X]) == {T: X, S: X}  # type: ignore
    assert LLMStructuredPrompt.get_generic_type_map(GenericType[U, U][X]) == {T: X, S: X}  # type: ignore
    assert LLMStructuredPrompt.get_generic_type_map(GenericType[int, U][str]) == {T: int, S: str}  # type: ignore


def test_resolve_generic_types() -> None:
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    # Generic Pydantic model
    class GenericType(GenericModel, typing.Generic[T, S]):
        a: T
        b: S

    assert LLMStructuredPrompt.resolve_generic_types(GenericType, GenericType[int, str](a=1, b="Hello")) == {
        T: int,
        S: str,
    }


def test_llm_function():
    @llm_function
    def add(llm: BaseLLM, a: int, b: int) -> int:
        """
        Add two numbers.
        """
        raise NotImplementedError

    fake_llm = FakeLLM(
        texts=[
            "Add two numbers.\n\nThe input and output are formatted as a JSON interface that conforms to the JSON "
            'schemas below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a list of '
            'strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object {"foo": ['
            '"bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": ["bar", '
            '"baz"]}} is not well-formatted.\n\nHere is the input schema:\n```\n{"properties": {"a": {"type": '
            '"integer"}, "b": {"type": "integer"}}, "required": ["a", "b"]}\n```\n\nHere is the output '
            'schema:\n```\n{"properties": {"return_value": {"type": "integer"}}, "required": ['
            '"return_value"]}\n```\nNow output the results for the following inputs:\n```\n{"a": 1, "b": 2}\n```'
            '{"return_value": 3}'
        ]
    )

    assert add.__doc__ == "\n        Add two numbers.\n        "
    assert add(fake_llm, 1, 2) == 3
    assert add.__doc__ == inspect.unwrap(add).__doc__



from langchain.schema import AIMessage, HumanMessage

from llm_strategy import adapters
from llm_strategy.testing import fake_chat_model, fake_llm


def test_chat_model_as_llm():
    chat_model = fake_chat_model.FakeChatModel.from_messages(
        [[HumanMessage(content="Hello", additional_kwargs={}), AIMessage(content="World", additional_kwargs={})]]
    )

    chat_model_as_llm = adapters.ChatModelAsLLM(chat_model=chat_model)

    assert chat_model_as_llm("Hello") == "World"


def test_llm_as_chat_model():
    llm = fake_llm.FakeLLM(texts={"<|im_start|>user\nHello<|im_end|><|im_start|>assistant\nWorld<|im_end|>"})

    chat_model_as_llm = adapters.LLMAsChatModel(llm=llm)

    assert chat_model_as_llm([HumanMessage(content="Hello")]) == AIMessage(content="World")



import langchain
import pytest

from llm_strategy.testing import fake_llm

langchain.llm_cache = None


def test_fake_llm_query():
    """Test that the fake LLM returns the correct query."""
    llm = fake_llm.FakeLLM(texts={"foobar"})
    assert llm("foo") == "bar"


def test_fake_llm_query_with_stop():
    """Test that the fake LLM returns the correct query."""
    llm = fake_llm.FakeLLM(texts={"foobar"})
    assert llm("foo", stop=["a"]) == "b"


def test_fake_llm_missing_query():
    """Test that the fake LLM raises an error if the query is missing."""
    llm = fake_llm.FakeLLM(texts=set())
    with pytest.raises(NotImplementedError):
        raise ValueError(llm("foo"))



from typing import Any, Mapping

from langchain.llms.base import LLM, BaseLLM
from pydantic import BaseModel, Field


class FakeLLM(LLM, BaseModel):
    """Fake LLM wrapper for testing purposes.

    We can use this to test the behavior of the LLM wrapper without having to actually call the LLM.

    We support an `external_llm` argument, which is an LLM that will be called if the query is not found in the `texts`
    dict. We store the responses. On exit, we deduplicate them and print them to stdout so that they can be copied into
    constructor call for the next run by hand if needed.

    We support stop words, which are words that are removed from the response if they are found. To do so, we store
    the full response (as it is build over time) and return the part before the query and the stop word.

    This also means that there is no non-determinism in the output, which is good for testing, but bad for variance.
    Especially if we want to test the behavior of the LLM wrapper when the LLM is not deterministic. (Create different
    outputs for different calls, for example.)
    """

    texts: set[str] = Field(default_factory=set)
    """The texts to return on call."""
    external_llm: BaseLLM | None = None
    """An external LLM to use if the query is not found."""

    def __del__(self) -> None:
        # If we have an external LLM, we write out all our responses to stdout so that they can be copied into the
        # constructor call for the next run by hand if needed.
        if self.external_llm is not None:
            # Deduplicate the texts (any shared prefixes can be removed)
            self.texts = {
                text for text in self.texts if not any(other.startswith(text) for other in self.texts if other != text)
            }
            print(f"texts = {self.texts!r}")

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "fake"

    def _call(self, prompt: str, stop: list[str] | None = None) -> str:
        """Return the query if it exists, else print the code to update the query."""
        for text in self.texts:
            if text.startswith(prompt):
                # Remainder:
                response = text[len(prompt) :]

                # Emulate stop behavior
                if stop is not None:
                    for stop_word in stop:
                        if stop_word in response:
                            # Only return the answer up to the stop word
                            response = response[: response.index(stop_word)]
                return response

        if self.external_llm is not None:
            response = self.external_llm.invoke(prompt, stop=stop)
            text = prompt + response
            self.texts.add(text)
            return response

        # If no queries are provided, print the code to update the query
        code_snippet = (
            f"# Add the following to the queries list:\n\n{repr(prompt)}\n# TODO: Append the correct response here"
        )
        print(code_snippet)
        raise NotImplementedError("No query provided to FakeLLM." + code_snippet)

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        return {}



#  Blackboard-PAGI - LLM Proto-AGI using the Blackboard Pattern
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest
from langchain.schema import AIMessage, HumanMessage, SystemMessage

from llm_strategy.testing import fake_chat_model


def test_fake_chat_model_query():
    """Test that the fake LLM returns the correct query."""
    chat_model = fake_chat_model.FakeChatModel.from_messages(
        [[SystemMessage(content="foo"), HumanMessage(content="bar"), AIMessage(content="doo")]]
    )
    assert chat_model([SystemMessage(content="foo"), HumanMessage(content="bar")]) == AIMessage(content="doo")


def test_fake_chat_model_query_with_stop_raise():
    """Test that the fake LLM returns the correct query."""
    chat_model = fake_chat_model.FakeChatModel.from_messages(
        [[SystemMessage(content="foo"), HumanMessage(content="bar"), AIMessage(content="doo")]]
    )

    with pytest.raises(AssertionError):
        chat_model([SystemMessage(content="foo"), HumanMessage(content="bar")], stop=["a"])


def test_chat_model_llm_missing_query():
    """Test that the fake LLM raises an error if the query is missing."""
    chat_model = fake_chat_model.FakeChatModel(messages_tuples_bag=set())
    with pytest.raises(NotImplementedError):
        chat_model([SystemMessage(content="foo"), HumanMessage(content="bar")])



# type: ignore
import dataclasses
import functools  # noqa: F401
import inspect
import typing
from dataclasses import dataclass

import typing_extensions
from langchain.llms.base import BaseLLM

from llm_strategy.llm_function import (
    LLMFunction,
    apply_decorator,
    is_not_implemented,
    unwrap_function,
)

P = typing_extensions.ParamSpec("P")
T = typing.TypeVar("T")


def can_wrap_function_in_llm(f: typing.Callable[P, T]) -> bool:
    """
    Return True if f can be wrapped in an LLMCall.
    """
    unwrapped = unwrap_function(f)
    if not callable(unwrapped):
        return False
    if inspect.isgeneratorfunction(unwrapped):
        return False
    if inspect.iscoroutinefunction(unwrapped):
        return False
    if not inspect.isfunction(unwrapped) and not inspect.ismethod(unwrapped):
        return False
    return is_not_implemented(unwrapped)


def llm_strategy(llm: BaseLLM) -> typing.Callable[[T], T]:  # noqa: C901
    """
    A strategy that implements whatever it decorates (or is called on) using the LLM.
    """

    @typing.no_type_check
    def decorator(f: T) -> T:
        assert can_wrap_member_in_llm(f)

        # For an instance of dataclass, call llm_strategy_dataclass with the fields.
        if dataclasses.is_dataclass(f):
            if isinstance(f, type):
                return llm_dataclass(f, llm)
            else:
                implemented_dataclass = llm_dataclass(type(f), llm)
                # Create an instance of the implemented dataclass using the fields from f
                params = {field.name: getattr(f, field.name) for field in dataclasses.fields(f)}
                return implemented_dataclass(**params)
        else:

            def inner_decorator(unwrapped_f):
                llm_f = None

                @functools.wraps(unwrapped_f)
                def strategy_wrapper(*args, **kwargs):
                    nonlocal llm_f
                    if llm_f is None:
                        # Get the signature of f
                        sig = inspect.signature(unwrapped_f, eval_str=True)
                        # Add a llm parameter to the signature as first argument
                        new_params = [inspect.Parameter("__llm", inspect.Parameter.POSITIONAL_ONLY)]
                        new_params.extend(sig.parameters.values())

                        new_sig = sig.replace(parameters=new_params)

                        def dummy_f(*args, **kwargs):
                            raise NotImplementedError()

                        new_f = functools.wraps(unwrapped_f)(dummy_f)
                        new_f.__module__ = unwrapped_f.__module__
                        # Set the signature of the new function
                        new_f.__signature__ = new_sig

                        del new_f.__wrapped__

                        # Wrap the function in an LLMFunction
                        llm_f = functools.wraps(new_f)(LLMFunction())

                    return llm_f(llm, *args, **kwargs)

                return strategy_wrapper

            return apply_decorator(f, inner_decorator)

    return decorator


def can_wrap_member_in_llm(f: typing.Callable[P, T]) -> bool:
    """
    Return True if f can be wrapped in an LLMCall.
    """
    if isinstance(f, LLMFunction):
        return True
    if dataclasses.is_dataclass(f):
        return True

    return can_wrap_function_in_llm(f)


@typing_extensions.dataclass_transform()
def llm_dataclass(dataclass_type: type, llm: BaseLLM) -> type:
    global long_unlikely_prefix__llm
    global long_unlikely__member_name, long_unlikely__member
    long_unlikely_prefix__llm = llm
    long_unlikely__member_name, long_unlikely__member = None, None

    @dataclass
    class SpecificLLMImplementation(dataclass_type):
        global long_unlikely__member_name, long_unlikely__member
        for long_unlikely__member_name, long_unlikely__member in inspect.getmembers_static(  # noqa: B007
            dataclass_type, can_wrap_member_in_llm
        ):
            exec(
                f"""
@llm_strategy(long_unlikely_prefix__llm)
@functools.wraps(long_unlikely__member)
def {long_unlikely__member_name}(*args, **kwargs):
    raise NotImplementedError()
"""
            )

    SpecificLLMImplementation.__name__ = f"{llm.__class__.__name__}_{dataclass_type.__name__}"
    SpecificLLMImplementation.__qualname__ = f"{llm.__class__.__name__}_{dataclass_type.__qualname__}"

    del long_unlikely__member_name, long_unlikely__member
    del long_unlikely_prefix__llm

    return SpecificLLMImplementation



from dataclasses import dataclass

from llm_strategy import llm_strategy
from llm_strategy.llm_function import is_not_implemented
from llm_strategy.testing.fake_llm import FakeLLM


@dataclass
class DummyInterface:
    """A dataclass with some members that are not implemented."""

    a: int = -1

    @staticmethod
    def static_method(a: int, b: int) -> int:
        """Add two numbers."""
        raise NotImplementedError()

    # @classmethod
    # def class_method(cls: type, a: int, b: int) -> int:
    #     """Add two numbers."""
    #     raise NotImplementedError()

    @property
    def property_getter(self: "DummyInterface") -> int:
        """Return 0."""
        raise NotImplementedError()

    def bound_method(self: "DummyInterface", a: int, b: int) -> int:
        """Add two numbers."""
        raise NotImplementedError()

    def bound_method_implemented(self: "DummyInterface", b: int = 1) -> int:
        """Add two numbers."""
        return self.a + b

    @staticmethod
    def static_method_implemented(a: int, b: int = 1) -> int:
        """Add two numbers."""
        return a + b

    @classmethod
    def class_method_implemented(cls, a: int, b: int = 1) -> int:
        """Add two numbers."""
        return a + b

    @property
    def property_getter_implemented(self) -> int:
        """Return 1."""
        return 1


def not_implemented_function():
    raise NotImplementedError


def test_is_not_implemented_functon():
    assert is_not_implemented(not_implemented_function)
    assert not is_not_implemented(lambda: 1)


def test_llm_strategy_on_functions():
    def add_two_ints(a: int, b: int) -> int:
        """Add two integers."""
        raise NotImplementedError()

    def add_two_ints_with_default(a: int, b: int = 1) -> int:
        """Add two integers with a default value."""
        raise NotImplementedError()

    def add_two_ints_with_default_and_kwarg(*, a: int, c: int = 2) -> int:
        """Add two integers with a default value."""
        raise NotImplementedError()

    llm = FakeLLM(
        texts={
            (
                "Add two integers.\n\nThe input and output are formatted as a JSON interface that conforms to the "
                'JSON schemas below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a list '
                'of strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object {'
                '"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": '
                '["bar", "baz"]}} is not well-formatted.\n\nHere is the input schema:\n```\n{"properties": {"a": {'
                '"type": "integer"}, "b": {"type": "integer"}}, "required": ["a", "b"]}\n```\n\nHere is the output '
                'schema:\n```\n{"properties": {"return_value": {"type": "integer"}}, "required": ['
                '"return_value"]}\n```\nNow output the results for the following inputs:\n```\n{"a": 1, "b": 2}\n```'
                '{"return_value": 3}\n'
            ),
            (
                "Add two integers with a default value.\n\nThe input and output are formatted as a JSON interface "
                'that conforms to the JSON schemas below.\n\nAs an example, for the schema {"properties": {"foo": {'
                '"description": "a list of strings", "type": "array", "items": {"type": "string"}}}, "required": ['
                '"foo"]}} the object {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {'
                '"properties": {"foo": ["bar", "baz"]}} is not well-formatted.\n\nHere is the input schema:\n```\n{'
                '"properties": {"a": {"type": "integer"}, "b": {"type": "integer"}}, "required": ["a"]}\n```\n\nHere '
                'is the output schema:\n```\n{"properties": {"return_value": {"type": "integer"}}, "required": ['
                '"return_value"]}\n```\nNow output the results for the following inputs:\n```\n{"a": 1, "b": 1}\n```'
                '{"return_value": 2}\n'
            ),
            (
                "Add two integers with a default value.\n\nThe input and output are formatted as a JSON interface "
                'that conforms to the JSON schemas below.\n\nAs an example, for the schema {"properties": {"foo": {'
                '"description": "a list of strings", "type": "array", "items": {"type": "string"}}}, "required": ['
                '"foo"]}} the object {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {'
                '"properties": {"foo": ["bar", "baz"]}} is not well-formatted.\n\nHere is the input schema:\n```\n{'
                '"properties": {"a": {"type": "integer"}, "c": {"type": "integer"}}, "required": ["a"]}\n```\n\nHere '
                'is the output schema:\n```\n{"properties": {"return_value": {"type": "integer"}}, "required": ['
                '"return_value"]}\n```\nNow output the results for the following inputs:\n```\n{"a": 1, "c": 2}\n```'
                '{"return_value": 3}\n'
            ),
        },
        # external_llm=OpenAI(),
    )

    assert llm_strategy(llm)(add_two_ints)(1, 2) == 3
    assert llm_strategy(llm)(add_two_ints_with_default)(1) == 2
    assert llm_strategy(llm)(add_two_ints_with_default_and_kwarg)(a=1) == 3


def test_llm_strategy_dummy_interface():
    llm = FakeLLM(
        texts=[
            (
                "Return 0.\n\nThe input and output are formatted as a JSON interface that conforms to the JSON schemas"
                ' below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a list of strings",'
                ' "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object {"foo": ["bar",'
                ' "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": ["bar",'
                ' "baz"]}} is not well-formatted.\n\nHere is the schema for additional data'
                ' types:\n```\n{"DummyInterface": {"properties": {"a": {"type": "integer"}}}}\n```\n\nHere is the input'
                ' schema:\n```\n{"properties": {"self": {"$ref": "DummyInterface"}}, "required": ["self"]}\n```\n\nHere'
                ' is the output schema:\n```\n{"properties": {"return_value": {"type": "integer"}}, "required":'
                ' ["return_value"]}\n```\nNow output the results for the following inputs:\n```\n{"self": {"a":'
                ' -1}}\n```{"return_value": 0}\n'
            ),
            (
                "Add two numbers.\n\nThe input and output are formatted as a JSON interface that conforms to the JSON"
                ' schemas below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a list of'
                ' strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object {"foo":'
                ' ["bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": ["bar",'
                ' "baz"]}} is not well-formatted.\n\nHere is the schema for additional data'
                ' types:\n```\n{"DummyInterface": {"properties": {"a": {"type": "integer"}}}}\n```\n\nHere is the input'
                ' schema:\n```\n{"properties": {"self": {"$ref": "DummyInterface"}, "a": {"type": "integer"}, "b":'
                ' {"type": "integer"}}, "required": ["self", "a", "b"]}\n```\n\nHere is the output'
                ' schema:\n```\n{"properties": {"return_value": {"type": "integer"}}, "required":'
                ' ["return_value"]}\n```\nNow output the results for the following inputs:\n```\n{"self": {"a": -1},'
                ' "a": 1, "b": 2}\n```{"return_value": 3}\n'
            ),
            (
                "Add two numbers.\n\nThe input is formatted as a JSON interface of Inputs that conforms to the JSON "
                "schema below, and the output should be formatted as a JSON instance of Outputs that conforms to the "
                'JSON schema below.\n\nAs an example, for the schema {"properties": {"foo": {"title": "Foo", '
                '"description": "a list of strings", "type": "array", "items": {"type": "string"}}}, "required": ['
                '"foo"]}} the object {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {'
                '"properties": {"foo": ["bar", "baz"]}} is not well-formatted.\n\nHere is the schema:\n```\n{'
                '"DummyInterface": {"properties": {"a": {"title": "A", "default": -1, "type": "integer"}}}, '
                '"Inputs": {"properties": {"self": {"$ref": "#/definitions/DummyInterface"}, "a": {"title": "A", '
                '"type": "integer"}, "b": {"title": "B", "type": "integer"}}, "required": ["self", "a", "b"]}, '
                '"Outputs": {"properties": {"return_value": {"title": "Return Value", "type": "integer"}}, '
                '"required": ["return_value"]}}\n```\n\nNow output the results for the following inputs:\n```\n{'
                '"self": {"a": -1}, "a": 1, "b": 2}\n```\n '
                '{"return_value": 3}\n'
            ),
            (
                "Add two numbers.\n\nThe input and output are formatted as a JSON interface that conforms to the JSON "
                'schemas below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a list of '
                'strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object {"foo": ['
                '"bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": ["bar", '
                '"baz"]}} is not well-formatted.\n\nHere is the input schema:\n```\n{"properties": {"a": {"type": '
                '"integer"}, "b": {"type": "integer"}}, "required": ["a", "b"]}\n```\n\nHere is the output '
                'schema:\n```\n{"properties": {"return_value": {"type": "integer"}}, "required": ['
                '"return_value"]}\n```\nNow output the results for the following inputs:\n```\n{"a": 1, "b": 2}\n```'
                '{"return_value": 3}\n'
            ),
        ]
        # external_llm=OpenAI(),
    )

    llm_DummyInterface = llm_strategy(llm)(DummyInterface)
    llm_DummyInstance = llm_DummyInterface()
    assert llm_DummyInstance.property_getter == 0
    assert llm_DummyInterface.static_method(1, 2) == 3
    assert llm_DummyInstance.static_method(1, 2) == 3
    assert llm_DummyInstance.bound_method(1, 2) == 3
    # assert llm_DummyInterface.class_method(1, 2) == 3
    # assert llm_DummyInstance.class_method(1, 2) == 3

    assert llm_DummyInterface.static_method_implemented(1, 2) == 3
    assert llm_DummyInstance.bound_method_implemented(1) == 0
    assert llm_DummyInterface.class_method_implemented(1, 2) == 3
    assert llm_DummyInstance.property_getter_implemented == 1


@dataclass
class Customer:
    first_name: str
    last_name: str
    birthday: str
    city: str


@dataclass
class CustomerDatabase:
    customers: list[Customer]

    def find_customer_index(self: "CustomerDatabase", query: str) -> int:
        """Find the index of the customer that matches the natural language query best.

        We support semantic queries instead of SQL, so we can search for things like
        "the customer that was born in 1990".

        Args:
            query: Natural language query

        Returns:
            The index of the best matching customer in the database.
        """
        raise NotImplementedError()

    @staticmethod
    def create_mock_customers(num_customers: int = 1) -> list[Customer]:
        """
        Create mock customers with believable data (our customers are world citizens).
        """
        raise NotImplementedError()

    @staticmethod
    def create_mock_queries(customer_database: "CustomerDatabase", num_queries: int = 1) -> list[str]:
        """
        Create mock queries that match one of the mock customers better than the others.

        We support semantic queries instead of SQL, so we can search for things like
        "the customer that was born in 1990".
        """
        raise NotImplementedError()


def test_llm_strategy():
    llm = FakeLLM(
        texts={
            (
                "Create mock queries that match one of the mock customers better than the others.\n\nWe support"
                ' semantic queries instead of SQL, so we can search for things like\n"the customer that was born in'
                ' 1990".\n\nThe input and output are formatted as a JSON interface that conforms to the JSON schemas'
                ' below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a list of strings",'
                ' "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object {"foo": ["bar",'
                ' "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo": ["bar",'
                ' "baz"]}} is not well-formatted.\n\nHere is the schema for additional data types:\n```\n{"Customer":'
                ' {"properties": {"first_name": {"type": "string"}, "last_name": {"type": "string"}, "birthday":'
                ' {"type": "string"}, "city": {"type": "string"}}, "required": ["first_name", "last_name", "birthday",'
                ' "city"]}, "CustomerDatabase": {"properties": {"customers": {"type": "array", "items": {"$ref":'
                ' "Customer"}}}, "required": ["customers"]}}\n```\n\nHere is the input schema:\n```\n{"properties":'
                ' {"customer_database": {"$ref": "CustomerDatabase"}, "num_queries": {"type": "integer"}}, "required":'
                ' ["customer_database"]}\n```\n\nHere is the output schema:\n```\n{"properties": {"return_value":'
                ' {"type": "array", "items": {"type": "string"}}}, "required": ["return_value"]}\n```\nNow output the'
                ' results for the following inputs:\n```\n{"customer_database": {"customers": [{"first_name": "John",'
                ' "last_name": "Smith", "birthday": "01/01/1990", "city": "New York"}, {"first_name": "Jane",'
                ' "last_name": "Doe", "birthday": "02/02/1995", "city": "London"}, {"first_name": "Alex", "last_name":'
                ' "Parker", "birthday": "03/03/1985", "city": "Paris"}]}, "num_queries": 3}\n```{"return_value": ["the'
                ' customer that was born in 1990", "the customer that lives in London", "the customer that has the last'
                ' name Parker"]}\n```'
            ),
            (
                "Create mock customers with believable data (our customers are world citizens).\n\nThe input and output"
                " are formatted as a JSON interface that conforms to the JSON schemas below.\n\nAs an example, for the"
                ' schema {"properties": {"foo": {"description": "a list of strings", "type": "array", "items": {"type":'
                ' "string"}}}, "required": ["foo"]}} the object {"foo": ["bar", "baz"]} is a well-formatted instance of'
                ' the schema. The object {"properties": {"foo": ["bar", "baz"]}} is not well-formatted.\n\nHere is the'
                ' schema for additional data types:\n```\n{"Customer": {"properties": {"first_name": {"type":'
                ' "string"}, "last_name": {"type": "string"}, "birthday": {"type": "string"}, "city": {"type":'
                ' "string"}}, "required": ["first_name", "last_name", "birthday", "city"]}}\n```\n\nHere is the input'
                ' schema:\n```\n{"properties": {"num_customers": {"type": "integer"}}}\n```\n\nHere is the output'
                ' schema:\n```\n{"properties": {"return_value": {"type": "array", "items": {"$ref": "Customer"}}},'
                ' "required": ["return_value"]}\n```\nNow output the results for the following'
                ' inputs:\n```\n{"num_customers": 3}\n```\n\nOutput:\n```\n{"return_value": [{"first_name": "John",'
                ' "last_name": "Smith", "birthday": "01/01/1990", "city": "New York"}, \n{"first_name": "Jane",'
                ' "last_name": "Doe", "birthday": "02/02/1995", "city": "London"}, \n{"first_name": "Alex",'
                ' "last_name": "Parker", "birthday": "03/03/1985", "city": "Paris"}]}'
            ),
            (
                "Find the index of the customer that matches the natural language query best.\n\nWe support semantic"
                ' queries instead of SQL, so we can search for things like\n"the customer that was born in'
                ' 1990".\n\nArgs:\n    query: Natural language query\n\nReturns:\n    The index of the best matching'
                " customer in the database.\n\nThe input and output are formatted as a JSON interface that conforms to"
                ' the JSON schemas below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a'
                ' list of strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object'
                ' {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo":'
                ' ["bar", "baz"]}} is not well-formatted.\n\nHere is the schema for additional data'
                ' types:\n```\n{"Customer": {"properties": {"first_name": {"type": "string"}, "last_name": {"type":'
                ' "string"}, "birthday": {"type": "string"}, "city": {"type": "string"}}, "required": ["first_name",'
                ' "last_name", "birthday", "city"]}, "CustomerDatabase": {"properties": {"customers": {"type": "array",'
                ' "items": {"$ref": "Customer"}}}, "required": ["customers"]}}\n```\n\nHere is the input'
                ' schema:\n```\n{"properties": {"self": {"$ref": "CustomerDatabase"}, "query": {"type": "string"}},'
                ' "required": ["self", "query"]}\n```\n\nHere is the output schema:\n```\n{"properties":'
                ' {"return_value": {"type": "integer"}}, "required": ["return_value"]}\n```\nNow output the results for'
                ' the following inputs:\n```\n{"self": {"customers": [{"first_name": "John", "last_name": "Smith",'
                ' "birthday": "01/01/1990", "city": "New York"}, {"first_name": "Jane", "last_name": "Doe", "birthday":'
                ' "02/02/1995", "city": "London"}, {"first_name": "Alex", "last_name": "Parker", "birthday":'
                ' "03/03/1985", "city": "Paris"}]}, "query": "the customer that was born in 1990"}\n```Output:'
                ' {"return_value": 2}'
            ),
            (
                "Find the index of the customer that matches the natural language query best.\n\nWe support semantic"
                ' queries instead of SQL, so we can search for things like\n"the customer that was born in'
                ' 1990".\n\nArgs:\n    query: Natural language query\n\nReturns:\n    The index of the best matching'
                " customer in the database.\n\nThe input and output are formatted as a JSON interface that conforms to"
                ' the JSON schemas below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a'
                ' list of strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object'
                ' {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo":'
                ' ["bar", "baz"]}} is not well-formatted.\n\nHere is the schema for additional data'
                ' types:\n```\n{"Customer": {"properties": {"first_name": {"type": "string"}, "last_name": {"type":'
                ' "string"}, "birthday": {"type": "string"}, "city": {"type": "string"}}, "required": ["first_name",'
                ' "last_name", "birthday", "city"]}, "CustomerDatabase": {"properties": {"customers": {"type": "array",'
                ' "items": {"$ref": "Customer"}}}, "required": ["customers"]}}\n```\n\nHere is the input'
                ' schema:\n```\n{"properties": {"self": {"$ref": "CustomerDatabase"}, "query": {"type": "string"}},'
                ' "required": ["self", "query"]}\n```\n\nHere is the output schema:\n```\n{"properties":'
                ' {"return_value": {"type": "integer"}}, "required": ["return_value"]}\n```\nNow output the results for'
                ' the following inputs:\n```\n{"self": {"customers": [{"first_name": "John", "last_name": "Smith",'
                ' "birthday": "01/01/1990", "city": "New York"}, {"first_name": "Jane", "last_name": "Doe", "birthday":'
                ' "02/02/1995", "city": "London"}, {"first_name": "Alex", "last_name": "Parker", "birthday":'
                ' "03/03/1985", "city": "Paris"}]}, "query": "the customer that lives in'
                ' London"}\n```Output:\n{"return_value": 0}'
            ),
            (
                "Find the index of the customer that matches the natural language query best.\n\nWe support semantic"
                ' queries instead of SQL, so we can search for things like\n"the customer that was born in'
                ' 1990".\n\nArgs:\n    query: Natural language query\n\nReturns:\n    The index of the best matching'
                " customer in the database.\n\nThe input and output are formatted as a JSON interface that conforms to"
                ' the JSON schemas below.\n\nAs an example, for the schema {"properties": {"foo": {"description": "a'
                ' list of strings", "type": "array", "items": {"type": "string"}}}, "required": ["foo"]}} the object'
                ' {"foo": ["bar", "baz"]} is a well-formatted instance of the schema. The object {"properties": {"foo":'
                ' ["bar", "baz"]}} is not well-formatted.\n\nHere is the schema for additional data'
                ' types:\n```\n{"Customer": {"properties": {"first_name": {"type": "string"}, "last_name": {"type":'
                ' "string"}, "birthday": {"type": "string"}, "city": {"type": "string"}}, "required": ["first_name",'
                ' "last_name", "birthday", "city"]}, "CustomerDatabase": {"properties": {"customers": {"type": "array",'
                ' "items": {"$ref": "Customer"}}}, "required": ["customers"]}}\n```\n\nHere is the input'
                ' schema:\n```\n{"properties": {"self": {"$ref": "CustomerDatabase"}, "query": {"type": "string"}},'
                ' "required": ["self", "query"]}\n```\n\nHere is the output schema:\n```\n{"properties":'
                ' {"return_value": {"type": "integer"}}, "required": ["return_value"]}\n```\nNow output the results for'
                ' the following inputs:\n```\n{"self": {"customers": [{"first_name": "John", "last_name": "Smith",'
                ' "birthday": "01/01/1990", "city": "New York"}, {"first_name": "Jane", "last_name": "Doe", "birthday":'
                ' "02/02/1995", "city": "London"}, {"first_name": "Alex", "last_name": "Parker", "birthday":'
                ' "03/03/1985", "city": "Paris"}]}, "query": "the customer that has the last name'
                ' Parker"}\n```{"return_value": 1}'
            ),
        }
        #
        # external_llm=OpenAI(),
    )

    LLMCustomerDatabase = llm_strategy(llm)(CustomerDatabase)

    customers = LLMCustomerDatabase.create_mock_customers(3)

    customer_database = LLMCustomerDatabase(customers)

    mock_queries = LLMCustomerDatabase.create_mock_queries(customer_database, 3)

    print(customer_database)

    for query in mock_queries:
        index = customer_database.find_customer_index(query)
        print(query)
        print(customer_database.customers[index])
        assert index >= 0
        assert index < len(customers)



from llm_strategy.llm_function import (  # type: ignore
    llm_explicit_function,
    llm_function,
)
from llm_strategy.llm_strategy import llm_dataclass, llm_strategy  # type: ignore

__version__ = "2.1.1"



# Add a manual test that makes sure that we can use the json response format for OpenAI
import openai
from langchain_community.chat_models import ChatOpenAI

from llm_strategy.chat_chain import ChatChain

base_llm = ChatOpenAI(max_tokens=64, model_name="gpt-3.5-turbo-1106")
# %%

chain = ChatChain(base_llm, [])
try:
    response = chain.query("What is 1+1? Give your reasoning, too.", model_args=chain.enforce_json_response())
except openai.BadRequestError as e:
    assert e.status_code == 400
    assert "must contain the word 'json'" in e.message
else:
    raise AssertionError("Expected an exception")



import pytest

from llm_hyperparameters.track_hyperparameters import (
    Hyperparameter,
    hyperparameters_scope,
    track_hyperparameters,
)


def test_all():
    @track_hyperparameters
    def f():
        return Hyperparameter() @ 1

    @track_hyperparameters
    def g():
        return Hyperparameter() @ 2

    with hyperparameters_scope() as scope:
        assert f() == 1
        assert g() == 2

    scope.hyperparameters[f].hparam0 = 3

    with scope():
        assert f() == 3
        assert g() == 2


def test_no_scope():
    @track_hyperparameters
    def f():
        return Hyperparameter() @ 1

    with pytest.warns(UserWarning):
        assert f() == 1


def test_no_name():
    @track_hyperparameters
    def f():
        return Hyperparameter() @ 1

    with hyperparameters_scope() as scope:
        assert f() == 1

    scope.hyperparameters[f].hparam0 = 2

    with scope():
        assert f() == 2

    @track_hyperparameters
    def g():
        return Hyperparameter() @ "Hello" + Hyperparameter() @ "Hello"

    with hyperparameters_scope() as scope:
        assert g() == "HelloHello"

    scope.hyperparameters[g].hparam1 = "World"

    with scope():
        assert g() == "HelloWorld"


def test_with_name():
    @track_hyperparameters
    def f():
        return Hyperparameter("hello") @ 1

    with hyperparameters_scope() as scope:
        assert f() == 1

    scope.hyperparameters[f].hello = 2

    with scope():
        assert f() == 2

    @track_hyperparameters
    def g():
        return Hyperparameter("hello") @ "Hello" + Hyperparameter("hello") @ "Hello"

    with hyperparameters_scope() as scope:
        assert g() == "HelloHello"

    scope.hyperparameters[g].hello = "World"

    with scope():
        assert g() == "WorldWorld"


def test_nested():
    @track_hyperparameters
    def f():
        return Hyperparameter("hello") @ 1 + Hyperparameter("world") @ 2

    assert f() == 3

    @track_hyperparameters
    def g():
        return Hyperparameter("hello") @ 3 + f()

    with hyperparameters_scope() as scope:
        assert g() == 6

    scope.hyperparameters[g].hello = 4

    with scope():
        assert g() == 7

    scope.hyperparameters[f].hello = 5

    with scope():
        assert g() == 11



#  Blackboard-PAGI - LLM Proto-AGI using the Blackboard Pattern
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional

import langchain
from langchain.chat_models import ChatOpenAI
from langchain.schema import (
    AIMessage,
    BaseMessage,
    ChatGeneration,
    ChatResult,
    Generation,
)


class CachedChatOpenAI(ChatOpenAI):
    def _generate(self, messages: List[BaseMessage], stop: Optional[List[str]] = None) -> ChatResult:
        messages_prompt = repr(messages)
        if langchain.llm_cache:
            results = langchain.llm_cache.lookup(messages_prompt, self.model_name)
            if results:
                assert len(results) == 1
                result: Generation = results[0]
                chat_result = ChatResult(
                    generations=[ChatGeneration(message=AIMessage(content=result.text))],
                    llm_output=result.generation_info,
                )
                return chat_result
        chat_result = super()._generate(messages, stop)
        if langchain.llm_cache:
            assert len(chat_result.generations) == 1
            result = Generation(text=chat_result.generations[0].message.content, generation_info=chat_result.llm_output)
            langchain.llm_cache.update(messages_prompt, self.model_name, [result])
        return chat_result



# type: ignore
import dis
import functools
import inspect
import json
import re
import string
import types
import typing
from copy import deepcopy
from dataclasses import dataclass

import pydantic
import pydantic.schema
import typing_extensions
from langchain.chat_models.base import BaseChatModel
from langchain.schema import OutputParserException
from langchain_core.language_models import BaseLanguageModel, BaseLLM
from llmtracer import TraceNodeKind, trace_calls, update_event_properties, update_name
from llmtracer.trace_builder import slicer
from pydantic import BaseModel, ValidationError, create_model, generics
from pydantic.fields import FieldInfo, Undefined
from pydantic.generics import replace_types

from llm_hyperparameters.track_hyperparameters import (
    Hyperparameter,
    track_hyperparameters,
)
from llm_strategy.chat_chain import ChatChain

T = typing.TypeVar("T")
S = typing.TypeVar("S")
P = typing_extensions.ParamSpec("P")
B = typing.TypeVar("B", bound=BaseModel)
C = typing.TypeVar("C", bound=BaseModel)
F = typing.TypeVar("F", bound=typing.Callable)


def get_json_schema_hyperparameters(schema: dict):
    """
    Get the hyperparameters from a JSON schema recursively.

    The hyperparameters are all fields for keys with "title" or "description".
    """
    hyperparameters = {}
    for key, value in schema.items():
        if key == "description":
            hyperparameters[key] = value
        elif isinstance(value, dict):
            sub_hyperparameters = get_json_schema_hyperparameters(value)
            if sub_hyperparameters:
                hyperparameters[key] = sub_hyperparameters
    return hyperparameters


def update_json_schema_hyperparameters(schema: dict, hyperparameters: dict):
    """
    Nested merge of the schema dict with the hyperparameters dict.
    """
    for key, value in hyperparameters.items():
        if key in schema:
            if isinstance(value, dict):
                update_json_schema_hyperparameters(schema[key], value)
            else:
                schema[key] = value
        else:
            schema[key] = value


def unwrap_function(f: typing.Callable[P, T]) -> typing.Callable[P, T]:
    # is f a property?
    if isinstance(f, property):
        f = f.fget
    # is f a wrapped function?
    elif hasattr(f, "__wrapped__"):
        f = inspect.unwrap(f)
    elif inspect.ismethod(f):
        f = f.__func__
    else:
        return f

    return unwrap_function(f)


def is_not_implemented(f: typing.Callable) -> bool:
    """Check that a function only raises NotImplementedError."""
    unwrapped_f = unwrap_function(f)

    if not hasattr(unwrapped_f, "__code__"):
        raise ValueError(f"Cannot check whether {f} is implemented. Where is __code__?")

    # Inspect the opcodes
    code = unwrapped_f.__code__
    # Get the opcodes
    opcodes = list(dis.get_instructions(code))
    # Check that it only uses the following opcodes:
    # - RESUME
    # - LOAD_GLOBAL
    # - PRECALL
    # - CALL
    # - RAISE_VARARGS
    valid_opcodes = {
        "RESUME",
        "LOAD_GLOBAL",
        "PRECALL",
        "CALL",
        "RAISE_VARARGS",
    }
    # We allow at most a function of length len(valid_opcodes)
    if len(opcodes) > len(valid_opcodes):
        return False
    for opcode in opcodes:
        if opcode.opname not in valid_opcodes:
            return False
        # Check that the function only raises NotImplementedError
        if opcode.opname == "LOAD_GLOBAL" and opcode.argval != "NotImplementedError":
            return False
        if opcode.opname == "RAISE_VARARGS" and opcode.argval != 1:
            return False
        valid_opcodes.remove(opcode.opname)
    # Check that the function raises a NotImplementedError at the end.
    if opcodes[-1].opname != "RAISE_VARARGS":
        return False
    return True


class TyperWrapper(str):
    """
    A wrapper around a type that can be used to create a Pydantic model.

    This is used to support @classmethods.
    """

    @classmethod
    def __get_validators__(cls) -> typing.Iterator[typing.Callable]:
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def validate(cls, v: type) -> str:
        if not isinstance(v, type):
            raise TypeError("type required")
        return v.__qualname__


class Output(pydantic.generics.GenericModel, typing.Generic[T]):
    return_value: T


@dataclass
class LLMStructuredPrompt(typing.Generic[B, T]):
    """
    A structured prompt for a language model.
    """

    docstring: str
    input_type: type[B]
    output_type: type[Output[T]]
    return_annotation: T

    input: B

    @staticmethod
    def extract_from_definitions(definitions: dict, type_: type) -> dict:
        normalized_name = pydantic.schema.normalize_name(type_.__name__)

        sub_schema = definitions[normalized_name]
        del definitions[normalized_name]

        return sub_schema

    def get_json_schema(self, exclude_default: bool = True) -> dict:
        schema = pydantic.schema.schema([self.input_type, self.output_type], ref_template="{model}")
        definitions: dict = deepcopy(schema["definitions"])
        # remove title and type from each sub dict in the definitions
        for value in definitions.values():
            value.pop("title")
            value.pop("type")

            for property in value.get("properties", {}).values():
                property.pop("title", None)
                if exclude_default:
                    property.pop("default", None)

        input_schema = self.extract_from_definitions(definitions, self.input_type)
        output_schema = self.extract_from_definitions(definitions, self.output_type)

        schema = dict(
            input_schema=input_schema,
            output_schema=output_schema,
            additional_definitions=definitions,
        )
        return schema

    @staticmethod
    def create(docstring: str, input_type: type[B], return_annotation: T, input: B) -> "LLMStructuredPrompt[B, T]":
        """Create an LLMExplicitFunction."""

        # determine the return type
        # the return type can be a type annotation or an Annotated type with annotation being a FieldInfo
        if typing.get_origin(return_annotation) is typing.Annotated:
            return_info = typing.get_args(return_annotation)
        else:
            return_info = (return_annotation, ...)

        # resolve generic types
        generic_type_map = LLMStructuredPrompt.resolve_generic_types(input_type, input)
        return_type: type = LLMStructuredPrompt.resolve_type(return_info[0], generic_type_map)

        if return_type is types.NoneType:  # noqa: E721
            raise ValueError(f"Resolve return type {return_info[0]} is None! This would be a NOP.")

        return_info = (return_type, return_info[1])

        if typing.get_origin(return_annotation) is typing.Annotated:
            assert hasattr(return_annotation, "copy_with")
            resolved_return_annotation = return_annotation.copy_with([return_info[0]])
        else:
            resolved_return_annotation = return_info[0]

        # create the output model
        resolved_output_model_type = Output[return_type]  # noqa

        # resolve input_type
        resolved_input_type = LLMStructuredPrompt.resolve_type(input_type, generic_type_map)

        return LLMStructuredPrompt(
            docstring=docstring,
            input_type=resolved_input_type,
            output_type=resolved_output_model_type,
            return_annotation=resolved_return_annotation,
            input=input,
        )

    @staticmethod
    def resolve_type(source_type: type, generic_type_map: dict[type, type]) -> type:
        """
        Resolve a type using the generic type map.

        Supports Pydantic.GenericModel and typing.Generic.
        """
        if source_type in generic_type_map:
            source_type = generic_type_map[source_type]

        if isinstance(source_type, type) and issubclass(source_type, generics.GenericModel):
            base_generic_type = LLMStructuredPrompt.get_base_generic_type(source_type)
            generic_parameter_type_map = LLMStructuredPrompt.get_generic_type_map(source_type, base_generic_type)
            # forward step using the generic type map
            resolved_generic_type_map = {
                generic_type: generic_type_map.get(target_type, target_type)
                for generic_type, target_type in generic_parameter_type_map.items()
            }
            resolved_tuple = tuple(
                resolved_generic_type_map[generic_type] for generic_type in base_generic_type.__parameters__
            )
            source_type = base_generic_type[resolved_tuple]
        else:
            # we let Pydantic handle the rest
            source_type = replace_types(source_type, generic_type_map)

        return source_type

    @staticmethod
    def resolve_generic_types(model: type[BaseModel], instance: BaseModel) -> dict:
        """
        Resolves the generic types of a given model instance and returns a generic type map.

        Args:
            model (type[BaseModel]): The model type.
            instance (BaseModel): The instance of the model.

        Returns:
            dict: The generic type map.
        """
        generic_type_map: dict = {}

        for field_name, attr_value in list(instance):
            if field_name not in model.__annotations__:
                continue

            annotation = model.__annotations__[field_name]

            # if the annotation is an Annotated type, get the type annotation
            if typing.get_origin(annotation) is typing.Annotated:
                annotation = typing.get_args(annotation)[0]

            # if the annotation is a type var, resolve it into the generic type map
            if isinstance(annotation, typing.TypeVar):
                LLMStructuredPrompt.add_resolved_type(generic_type_map, annotation, type(attr_value))
            # if the annotation is a generic type alias ignore
            elif isinstance(annotation, types.GenericAlias):
                # The generic type alias is not supported yet
                # The problem is that GenericAlias types are elided: e.g. type(list[str](["hello"])) -> list and not list[str].
                # But either way, we would need to resolve the types based on the actual elements and their mros.
                continue
            # if the annotation is a type, check if it is a generic type
            elif isinstance(annotation, type) and issubclass(annotation, generics.GenericModel):
                # check if the type is in generics._assigned_parameters
                generic_definition_type_map = LLMStructuredPrompt.get_generic_type_map(annotation)

                argument_type = type(attr_value)
                generic_instance_type_map = LLMStructuredPrompt.get_generic_type_map(argument_type)

                assert list(generic_definition_type_map.keys()) == list(generic_instance_type_map.keys())

                # update the generic type map
                # if the generic type is already in the map, check that it is the same
                for generic_parameter, generic_parameter_target in generic_definition_type_map.items():
                    if generic_parameter_target not in annotation.__parameters__:
                        continue
                    resolved_type = generic_instance_type_map[generic_parameter]
                    LLMStructuredPrompt.add_resolved_type(generic_type_map, generic_parameter_target, resolved_type)
            else:
                # Let Pydantic handle the rest
                continue

        return generic_type_map

    @staticmethod
    def add_resolved_type(generic_type_map, source_type, resolved_type):
        """
        Add a resolved type to the generic type map.
        """
        if source_type in generic_type_map:
            # TODO: support finding the common base class?
            if (previous_resolution := generic_type_map[source_type]) is not resolved_type:
                raise ValueError(
                    f"Cannot resolve generic type {source_type}, conflicting "
                    f"resolution: {previous_resolution} and {resolved_type}."
                )
        else:
            generic_type_map[source_type] = resolved_type

    @staticmethod
    def get_generic_type_map(generic_type, base_generic_type=None):
        """Build a generic type map for a generic type.

        It maps the generic type variables to the actual types.
        """

        if base_generic_type is None:
            base_generic_type = LLMStructuredPrompt.get_base_generic_type(generic_type)

        base_classes = inspect.getmro(generic_type)
        # we have to iterate through the base classes
        generic_parameter_type_map = {generic_type: generic_type for generic_type in generic_type.__parameters__}
        for base_class in base_classes:
            # skip baseclasses that are from pydantic.generic
            # this avoids a bug that is caused by generics.GenericModel.__parameterized_bases_
            if base_class.__module__ == "pydantic.generics":
                continue
            if issubclass(base_class, base_generic_type):
                if base_class in generics._assigned_parameters:
                    assignment = generics._assigned_parameters[base_class]
                    generic_parameter_type_map = {
                        old_generic_type: generic_parameter_type_map.get(new_generic_type, new_generic_type)
                        for old_generic_type, new_generic_type in assignment.items()
                    }

        return generic_parameter_type_map

    @staticmethod
    def get_base_generic_type(field_type: type) -> type[generics.GenericModel]:
        """Determine the base generic type of a generic type. E.g. List[str] -> List.

        Args:
            field_type (type): The generic type.

        Raises:
            ValueError: If the base generic type cannot be found.

        Returns:
            type[generics.GenericModel]: The base generic type.
        """

        # get the base class name from annotation (which is without [])
        base_generic_name = field_type.__name__
        if "[" in field_type.__name__:
            base_generic_name = field_type.__name__.split("[")[0]
        # get the base class from argument_type_base_classes with base_generic_name
        for base_class in reversed(inspect.getmro(field_type)):
            if base_class.__name__ == base_generic_name and issubclass(field_type, base_class):
                base_generic_type = base_class
                break
        else:
            raise ValueError(f"Could not find base generic type {base_generic_name} for {field_type}.")
        return base_generic_type

    @trace_calls(name="LLMStructuredPrompt", kind=TraceNodeKind.CHAIN, capture_args=False, capture_return=False)
    def __call__(
        self,
        language_model_or_chat_chain: BaseLanguageModel | ChatChain,
    ) -> T:
        """Call the function."""

        # check that the first argument is an instance of BaseLanguageModel
        # or a TrackedChatChain or UntrackedChatChain
        if not isinstance(language_model_or_chat_chain, BaseLanguageModel | ChatChain):
            raise ValueError("The first parameter must be an instance of BaseLanguageModel or ChatChain.")

        # get the input and output schema as JSON dict
        schema = self.get_json_schema()
        # print(json.dumps(schema, indent=1))

        update_json_schema_hyperparameters(
            schema,
            Hyperparameter("json_schema") @ get_json_schema_hyperparameters(schema),
        )

        update_event_properties(
            dict(
                arguments=dict(self.input),
            )
        )

        parsed_output = self.query(language_model_or_chat_chain, schema)

        # print(f"Input: {self.input.json(indent=1)}")
        # print(f"Output: {json.dumps(json.loads(parsed_output.json())['return_value'], indent=1)}")

        update_event_properties(dict(result=parsed_output.return_value))

        return parsed_output.return_value

    @track_hyperparameters
    def query(self, language_model_or_chat_chain, schema):  # noqa: C901
        # create the prompt
        json_dumps_kwargs = Hyperparameter("json_dumps_kwargs") @ dict(indent=None)
        additional_definitions_prompt_template = Hyperparameter(
            "additional_definitions_prompt_template",
            "Here is the schema for additional data types:\n```\n{additional_definitions}\n```\n\n",
        )

        optional_additional_definitions_prompt = ""
        if schema["additional_definitions"]:
            optional_additional_definitions_prompt = additional_definitions_prompt_template.format(
                additional_definitions=json.dumps(schema["additional_definitions"], **json_dumps_kwargs)
            )

        prompt = (
            Hyperparameter(
                "llm_structured_prompt_template",
                description=(
                    "The general-purpose prompt for the structured prompt execution. It tells the LLM what to "
                    "do and how to read function arguments and structure return values. "
                ),
            )
            @ '{docstring}\n\nThe input and output are formatted as a JSON interface that conforms to the JSON schemas below.\n\nAs an example, for the schema {{"properties": {{"foo": {{"description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}}} the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.\n\n{optional_additional_definitions_prompt}Here is the input schema:\n```\n{input_schema}\n```\n\nHere is the output schema:\n```\n{output_schema}\n```\nNow output the results for the following inputs:\n```\n{inputs}\n```'
        ).format(
            docstring=self.docstring,
            optional_additional_definitions_prompt=optional_additional_definitions_prompt,
            input_schema=json.dumps(schema["input_schema"], **json_dumps_kwargs),
            output_schema=json.dumps(schema["output_schema"], **json_dumps_kwargs),
            inputs=self.input.json(**json_dumps_kwargs),
        )

        # get the response
        num_retries = Hyperparameter("num_retries_on_parser_failure") @ 3
        if language_model_or_chat_chain is None:
            raise ValueError("The language model or chat chain must be provided.")

        if isinstance(language_model_or_chat_chain, BaseChatModel):
            language_model_or_chat_chain = ChatChain(language_model_or_chat_chain, [])

        if isinstance(language_model_or_chat_chain, ChatChain):
            chain = language_model_or_chat_chain
            for _ in range(num_retries):
                output, chain = chain.query(prompt, model_args=chain.enforce_json_response())

                try:
                    parsed_output = parse(output, self.output_type)
                    break
                except OutputParserException as e:
                    prompt = (
                        Hyperparameter("error_prompt") @ "Tried to parse your output but failed:\n\n"
                        + str(e)
                        + Hyperparameter("retry_prompt") @ "\n\nPlease try again and avoid this issue."
                    )
            else:
                exception = OutputParserException(f"Failed to parse the output after {num_retries} retries.")
                exception.add_note(chain)
                raise exception
        elif isinstance(language_model_or_chat_chain, BaseLLM):
            model: BaseChatModel = language_model_or_chat_chain
            # Check if the language model is of type "openai" and extend model args with a response format in that case
            model_dict = model.dict()
            if "openai" in model_dict["_type"] and model_dict.get("model_name") in (
                "gpt-4-1106-preview",
                "gpt-3.5-turbo-1106",
            ):
                model_args = dict(response_format=dict(type="json_object"))
            else:
                model_args = {}

            for _ in range(num_retries):
                output = model(prompt, **model_args)

                try:
                    parsed_output = parse(output, self.output_type)
                    break
                except OutputParserException as e:
                    prompt = (
                        prompt
                        + Hyperparameter("output_prompt") @ "\n\nReceived the output\n\n"
                        + output
                        + Hyperparameter("error_prompt") @ "Tried to parse your output but failed:\n\n"
                        + str(e)
                        + Hyperparameter("retry_prompt") @ "\n\nPlease try again and avoid this issue."
                    )
            else:
                exception = OutputParserException(f"Failed to parse the output after {num_retries} retries.")
                exception.add_note(prompt)
                raise exception
        else:
            raise ValueError("The language model or chat chain must be provided.")
        return parsed_output


@dataclass
class LLMBoundSignature:
    """
    A function call that can be used to generate a prompt.
    """

    structured_prompt: LLMStructuredPrompt
    signature: inspect.Signature

    @property
    def input_type(self) -> type[BaseModel]:
        """Return the input type."""
        return self.structured_prompt.input_type

    @property
    def output_type(self) -> type[BaseModel]:
        """Return the output type."""
        return self.structured_prompt.output_type

    @property
    def docstring(self) -> str:
        """Return the docstring."""
        return self.structured_prompt.docstring

    @property
    def return_annotation(self) -> str:
        """Return the name."""
        return self.structured_prompt.return_annotation

    def get_input_object(self, *args: P.args, **kwargs: P.kwargs) -> BaseModel:
        """Call the function and return the inputs."""
        # bind the inputs to the signature
        bound_arguments = LLMBoundSignature.bind(self.signature, args, kwargs)
        # get the arguments
        arguments = bound_arguments.arguments
        inputs = self.structured_prompt.input_type(**arguments)
        return inputs

    @staticmethod
    def from_call(f: typing.Callable[P, T], args: P.args, kwargs: P.kwargs) -> "LLMBoundSignature":  # noqa: C901
        """Create an LLMBoundSignature from a function.

        Args:
            f: The function to create the LLMBoundSignature from.
            args: The positional arguments to the function (but excluding the language model/first param).
            kwargs: The keyword arguments to the function.

        """

        # get clean docstring
        docstring = Hyperparameter("docstring") @ inspect.getdoc(f)
        if docstring is None:
            raise ValueError("The function must have a docstring.")

        # get the type of the first argument
        globals_from_f = f.__globals__ if hasattr(f, "__globals__") else None
        locals_from_f = f.__dict__ if hasattr(f, "__dict__") else None
        signature = inspect.signature(f, eval_str=True, globals=globals_from_f, locals=locals_from_f)

        # get all parameters
        parameters_items: list[tuple[str, inspect.Parameter]] = list(signature.parameters.items())
        # check that there is at least one parameter
        if not parameters_items:
            raise ValueError("The function must have at least one parameter.")
        # check that the first parameter has a type annotation that is an instance of BaseLanguageModel
        # or a TrackedChatChain
        first_parameter: inspect.Parameter = parameters_items[0][1]
        if first_parameter.annotation is not inspect.Parameter.empty:
            if not issubclass(first_parameter.annotation, BaseLanguageModel | ChatChain):
                raise ValueError("The first parameter must be an instance of BaseLanguageModel or ChatChain.")

        return_type = signature.return_annotation
        if return_type is inspect.Parameter.empty:
            raise ValueError("The function must have a return type.")

        # create a pydantic model from the parameters
        parameter_dict = LLMBoundSignature.parameter_items_to_field_tuple(parameters_items[1:])

        # turn function name into a class name
        class_name = string.capwords(f.__name__, sep="_").replace("_", "")

        # create the input model
        # collect all __parameters__ from the type annotations
        # this is necessary because we need to know the types of the parameters
        # to create the pydantic model
        generic_parameters: set[typing.TypeVar] = set()
        for parameter in parameter_dict.values():
            annotation = parameter[0]
            # unwrap any Annotated types
            while hasattr(annotation, "__metadata__"):
                annotation = annotation.__origin__
            # if the annotation is already a type variable, add it to the set
            if isinstance(annotation, typing.TypeVar):
                generic_parameters.add(annotation)
            # if the annotation is a generic type, add the parameters to the set
            if hasattr(annotation, "__parameters__"):
                generic_parameters.update(annotation.__parameters__)

        model_spec = LLMBoundSignature.field_tuples_to_model_spec(parameter_dict)
        if generic_parameters:
            bases = (pydantic.generics.GenericModel, typing.Generic[*generic_parameters])
            input_type = create_model(f"{class_name}Inputs", __base__=bases, __module__=f.__module__, **model_spec)
        else:
            input_type = create_model(f"{class_name}Inputs", __module__=f.__module__, **model_spec)
        input_type.update_forward_refs()

        # update parameter_dict types with bound_arguments
        # this ensures that we serialize the actual types
        # might not be optimal because the language model won't be aware of original types, however

        bound_arguments = LLMBoundSignature.bind(signature, args, kwargs)
        for parameter_name in parameter_dict:
            if parameter_name in bound_arguments.arguments:
                parameter_dict[parameter_name] = (
                    type(bound_arguments.arguments[parameter_name]),
                    parameter_dict[parameter_name][1],
                )

        specific_model_spec = LLMBoundSignature.field_tuples_to_model_spec(parameter_dict)
        specific_input_type = create_model(
            f"Specific{class_name}Inputs", __module__=f.__module__, **specific_model_spec
        )
        specific_input_type.update_forward_refs()

        input = specific_input_type(**bound_arguments.arguments)

        llm_structured_prompt: LLMStructuredPrompt = LLMStructuredPrompt.create(
            docstring=docstring,
            input_type=input_type,
            return_annotation=return_type,
            input=input,
        )

        return LLMBoundSignature(llm_structured_prompt, signature)

    @staticmethod
    def parameter_items_to_field_tuple(parameters_items: list[tuple[str, inspect.Parameter]]):
        """
        Get the parameter definitions for a function call from the parameters and arguments.
        """
        parameter_dict: dict = {}
        for parameter_name, parameter in parameters_items:
            # every parameter must be annotated or have a default value
            annotation = parameter.annotation
            if annotation is type:
                annotation = TyperWrapper

            if parameter.default is inspect.Parameter.empty:
                parameter_dict[parameter_name] = (annotation, ...)
            else:
                parameter_dict[parameter_name] = (annotation, parameter.default)

        return parameter_dict

    @staticmethod
    def field_tuples_to_model_spec(
        field_tuples_dict: dict[str, tuple[str, tuple[type, ...]]]
    ) -> dict[str, tuple[type, object] | object]:
        """
        Get the parameter definitions for a function call from the parameters and arguments.
        """
        parameter_dict: dict = {}
        for parameter_name, (annotation, default) in field_tuples_dict.items():
            # every parameter must be annotated or have a default value
            if default is ...:
                parameter_dict[parameter_name] = (annotation, ...)
            else:
                if annotation is not inspect.Parameter.empty:
                    parameter_dict[parameter_name] = (annotation, default)
                else:
                    parameter_dict[parameter_name] = default

        return parameter_dict

    @staticmethod
    def get_or_create_pydantic_default(field: FieldInfo):
        if field.default is not Undefined:
            if field.default is Ellipsis:
                return inspect.Parameter.empty
            return field.default
        if field.default_factory is not None:
            return field.default_factory()
        return None

    @staticmethod
    def bind(signature, args, kwargs):
        """
        Bind function taking into account Field definitions and defaults.

        The first parameter from the original signature is dropped (as it is the language model or chat chain).
        args and kwargs are bound to the remaining parameters.
        """
        # resolve parameter defaults to FieldInfo.default if the parameter is a field
        signature_fixed_defaults = signature.replace(
            parameters=[
                parameter.replace(default=LLMBoundSignature.get_or_create_pydantic_default(parameter.default))
                if isinstance(parameter.default, FieldInfo)
                else parameter
                for parameter in list(signature.parameters.values())[1:]
            ]
        )
        bound_arguments = signature_fixed_defaults.bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        return bound_arguments


class LLMFunctionInterface(typing.Generic[P, T], typing.Callable[P, T]):
    def get_structured_prompt(self, *args: P.args, **kwargs: P.kwargs) -> LLMStructuredPrompt:
        raise NotImplementedError

    def llm_bound_signature(self, *args, **kwargs) -> LLMBoundSignature:
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class LLMFunction(LLMFunctionInterface[P, T], typing.Generic[P, T]):
    """
    A callable that can be called with a chat model.
    """

    def llm_bound_signature(self, *args, **kwargs) -> LLMBoundSignature:
        return LLMBoundSignature.from_call(self, args, kwargs)

    def get_input_object(self, *args, **kwargs) -> BaseModel:
        return self.llm_bound_signature(*args, **kwargs).get_input_object(*args, **kwargs)

    def __get__(self, instance: object, owner: type | None = None) -> typing.Callable:
        """Support instance methods."""
        if instance is None:
            return self

        # Bind self to instance as MethodType
        return types.MethodType(self, instance)

    def __getattr__(self, item):
        return getattr(self.__wrapped__, item)

    def explicit(self, language_model_or_chat_chain: BaseLanguageModel | ChatChain, input_object: BaseModel):
        """Call the function with explicit inputs."""

        return track_hyperparameters(self)(language_model_or_chat_chain, **dict(input_object))

    @trace_calls(kind=TraceNodeKind.CHAIN, capture_return=slicer[1:], capture_args=True)
    def __call__(
        self,
        language_model_or_chat_chain: BaseLanguageModel | ChatChain,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        """Call the function."""
        update_name(self.__name__)

        # check that the first argument is an instance of BaseLanguageModel
        # or a TrackedChatChain or UntrackedChatChain
        if not isinstance(language_model_or_chat_chain, BaseLanguageModel | ChatChain):
            raise ValueError("The first parameter must be an instance of BaseLanguageModel or ChatChain.")

        # We expect that we wrap a function that raises NotImplementedError
        # We call it, so we can set breakpoints in the function
        try:
            self.__wrapped__(language_model_or_chat_chain, *args, **kwargs)
            raise ValueError("The function must raise NotImplementedError.")
        except NotImplementedError:
            pass

        llm_bound_signature = LLMBoundSignature.from_call(self, args, kwargs)
        return_value = llm_bound_signature.structured_prompt(language_model_or_chat_chain)
        return return_value


class LLMExplicitFunction(LLMFunctionInterface[P, T], typing.Generic[P, T]):
    """
    A callable that can be called with a chat model.
    """

    def llm_bound_signature(self, input: BaseModel) -> LLMBoundSignature:
        """Create an LLMFunctionSpec from a function."""
        # get clean docstring of
        docstring = inspect.getdoc(self)
        if docstring is None:
            raise ValueError("The function must have a docstring.")

        # get the type of the first argument
        signature = inspect.signature(self, eval_str=True)

        # get all parameters
        parameters_items: list[tuple[str, inspect.Parameter]] = list(signature.parameters.items())
        # check that there is at least one parameter
        if not parameters_items:
            raise ValueError("The function must have at least one parameter.")
        # check that the first parameter has a type annotation that is an instance of BaseLanguageModel
        # or a TrackedChatChain
        first_parameter: inspect.Parameter = parameters_items[0][1]
        if first_parameter.annotation is not inspect.Parameter.empty:
            if not issubclass(first_parameter.annotation, BaseLanguageModel | ChatChain):
                raise ValueError("The first parameter must be an instance of BaseLanguageModel or ChatChain.")

        second_parameter: inspect.Parameter = parameters_items[1][1]

        llm_structured_prompt = LLMStructuredPrompt.create(
            docstring=docstring,
            input_type=second_parameter.annotation,
            return_annotation=signature.return_annotation,
            input=input,
        )
        return LLMBoundSignature(llm_structured_prompt, signature)

    def __get__(self, instance: object, owner: type | None = None) -> typing.Callable:
        """Support instance methods."""
        if instance is None:
            return self

        # Bind self to instance as MethodType
        return types.MethodType(self, instance)

    def __getattr__(self, item):
        return getattr(self.__wrapped__, item)

    @trace_calls(kind=TraceNodeKind.CHAIN, capture_return=True, capture_args=slicer[1:])
    def __call__(self, language_model_or_chat_chain: BaseLanguageModel | ChatChain, input: BaseModel) -> T:
        """Call the function."""
        update_name(self.__name__)

        # check that the first argument is an instance of BaseLanguageModel
        # or a TrackedChatChain or UntrackedChatChain
        if not isinstance(language_model_or_chat_chain, BaseLanguageModel | ChatChain):
            raise ValueError("The first parameter must be an instance of BaseLanguageModel or ChatChain.")

        # We expect that we wrap a function that raises NotImplementedError
        # We call it, so we can set breakpoints in the function
        try:
            self.__wrapped__(language_model_or_chat_chain, input)
            raise ValueError("The function must raise NotImplementedError.")
        except NotImplementedError:
            pass

        llm_bound_signature = self.llm_bound_signature(input)
        return_value = llm_bound_signature.structured_prompt(language_model_or_chat_chain)
        return return_value


F_types: typing.TypeAlias = (
    F
    | LLMFunction[P, T]
    | LLMExplicitFunction[P, T]
    | types.MethodType
    | types.FunctionType
    | types.ClassMethodDescriptorType
    | types.MethodDescriptorType
    | types.MemberDescriptorType
    | types.MethodWrapperType
    | LLMFunctionInterface
)


def apply_decorator(f: F_types, decorator) -> F_types:
    """
    Apply a decorator to a function.

    This function is used to apply a decorator to a function, while preserving the function type.
    This is useful when we want to apply a decorator to a function that is a classmethod, staticmethod, property,
    or a method of a class.

    Parameters
    ----------

    f: F_types
        The function to decorate.
    decorator: Callable
        The decorator to apply.

    Returns
    -------

    F_types
        The decorated function.

    Raises
    ------

    ValueError
        If the function is a classmethod, staticmethod, property, or a method of a class.
    """
    specific_llm_function: object
    if isinstance(f, classmethod):
        raise ValueError("Cannot decorate classmethod with llm_strategy (no translation of cls: type atm).")
    elif isinstance(f, staticmethod):
        specific_llm_function = staticmethod(apply_decorator(f.__func__, decorator))
    elif isinstance(f, property):
        specific_llm_function = property(apply_decorator(f.fget, decorator), doc=f.__doc__)
    elif isinstance(f, types.MethodType):
        specific_llm_function = types.MethodType(apply_decorator(f.__func__, decorator), f.__self__)
    elif hasattr(f, "__wrapped__"):
        return apply_decorator(f.__wrapped__, decorator)
    elif isinstance(f, LLMFunctionInterface):
        specific_llm_function = f
    elif not callable(f):
        raise ValueError(f"Cannot decorate {f} with llm_strategy.")
    else:
        if not is_not_implemented(f):
            raise ValueError("The function must not be implemented.")

        specific_llm_function = track_hyperparameters(functools.wraps(f)(decorator(f)))

    return typing.cast(F_types, specific_llm_function)


def llm_explicit_function(f: F_types) -> F_types:
    """
    Decorator to wrap a function with a chat model.

    f is a function to a dataclass or Pydantic model.

    The docstring of the function provides instructions for the model.
    """
    return apply_decorator(f, lambda f: LLMExplicitFunction())


def llm_function(f: F_types) -> F_types:
    """
    Decorator to wrap a function with a chat model.

    f is a function to a dataclass or Pydantic model.

    The docstring of the function provides instructions for the model.
    """
    return apply_decorator(f, lambda f: LLMFunction())


def parse(text: str, output_model: type[B]) -> B:
    try:
        # Greedy search for 1st json candidate.
        match = re.search(r"\{.*\}", text.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
        json_str = ""
        if match:
            json_str = match.group()
        json_object = json.loads(json_str)
        return output_model.parse_obj(json_object)

    except (json.JSONDecodeError, ValidationError) as e:
        msg = f'Failed to parse the last reply. Expected: `{{"return_value": ...}}` Got: {e}'
        raise OutputParserException(msg)


_typing_GenericAlias = type(typing.List[int])


# TODO: move this to the hyperparameter optimizer
def get_concise_type_repr(return_type: type):
    """Return a shorter (string) representation of the return type.

    Examples:

        <class 'str'> -> str
        <class 'int'> -> int
        <class 'CustomType'> -> CustomType
        <class 'typing.List[typing.Dict[str, int]]'> -> List[Dict[str, int]]

    For generic types, we want to keep the type arguments as well.

        <class 'typing.List[typing.Dict[str, int]]'> -> List[Dict[str, int]]
        <class 'PydanticGenericModel[typing.Dict[str, int]]'> -> PydanticGenericModel[Dict[str, int]]

    For unspecialized generic types, we want to keep the type arguments as well.

        so for class PydanticGenericModel(Generic[T]): pass:
            -> PydanticGenericModel[T]
    """
    assert isinstance(return_type, type | types.GenericAlias | _typing_GenericAlias | typing.TypeVar), return_type
    name = return_type.__name__
    # is it a specialized generic type?
    if hasattr(return_type, "__origin__"):
        origin = return_type.__origin__
        if origin is not None:
            # is it a generic type with type arguments?
            if hasattr(return_type, "__args__"):
                args = return_type.__args__
                if args:
                    # is it a generic type with type arguments?
                    args_str = ", ".join([get_concise_type_repr(arg) for arg in args])
                    return f"{origin.__name__}[{args_str}]"
    # is it a unspecialized generic type?
    if hasattr(return_type, "__parameters__"):
        parameters = return_type.__parameters__
        if parameters:
            # is it a generic type without type arguments?
            parameters_str = ", ".join([get_concise_type_repr(parameter) for parameter in parameters])
            return f"{name}[{parameters_str}]"

    return name



import dataclasses
import functools
import string
import typing
import warnings
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass

from pydantic import BaseModel, Field, create_model
from pydantic.fields import FieldInfo, Undefined, UndefinedType
from pydantic.typing import NoArgAnyCallable

from llm_hyperparameters.utils.callable_wrapper import CallableWrapper

if typing.TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny

T = typing.TypeVar("T")
P = typing.ParamSpec("P")


@dataclass
class HyperparameterDefinition:
    name: str
    explicit_type: type | UndefinedType
    field_info: FieldInfo

    def __matmul__(self, default: T) -> T:
        if _hyperparameter_context is not None:
            self.field_info.default = default
            return _hyperparameter_context.hyperparameter_builder.get_hyperparameter(self.name, default)
        return default


class IdentityDefinition:
    def __matmul__(self, default: T) -> T:
        return default


def Hyperparameter(
    name: str | None = None,
    default: T = Undefined,  # type: ignore
    *,
    description: str | None = None,
    explicit_type: type[T] = Undefined,  # type: ignore
    default_factory: NoArgAnyCallable | None = None,
    alias: str | None = None,
    title: str | None = None,
    exclude: typing.Union["AbstractSetIntStr", "MappingIntStrAny", typing.Any, None] = None,
    include: typing.Union["AbstractSetIntStr", "MappingIntStrAny", typing.Any, None] = None,
    const: bool | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    multiple_of: float | None = None,
    allow_inf_nan: bool | None = None,
    max_digits: int | None = None,
    decimal_places: int | None = None,
    min_items: int | None = None,
    max_items: int | None = None,
    unique_items: bool | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    allow_mutation: bool = True,
    regex: str | None = None,
    discriminator: str | None = None,
    repr: bool = True,
    **extra: typing.Any,
) -> T | HyperparameterDefinition | IdentityDefinition:  # type: ignore
    definition: HyperparameterDefinition | IdentityDefinition
    if _hyperparameter_context is not None:
        field = Field(
            default_factory=default_factory,  # type: ignore
            alias=alias,
            title=title,
            description=description,
            exclude=exclude,
            include=include,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            max_digits=max_digits,
            decimal_places=decimal_places,
            min_items=min_items,
            max_items=max_items,
            unique_items=unique_items,
            min_length=min_length,
            max_length=max_length,
            allow_mutation=allow_mutation,
            regex=regex,
            discriminator=discriminator,
            repr=repr,
            **extra,
        )

        if name is None:
            name = f"hparam{_hyperparameter_context.unique_id}"
            _hyperparameter_context.unique_id += 1

        definition = HyperparameterDefinition(name, explicit_type, field)

        _hyperparameter_context.hyperparameter_builder.hyperparameter_definitions[name] = definition
    else:
        definition = IdentityDefinition()
        # log warning
        warnings.warn(UserWarning(f"Hyperparameter {name or default} defined outside of a tracked function!"))

    if default is Undefined:
        # return a wrapper that supports matmul
        return definition
    else:
        return definition @ default


@dataclass
class HyperparameterBuilder:
    module: str
    qualname: str
    hyperparameter_definitions: dict[str, HyperparameterDefinition] = dataclasses.field(default_factory=dict)
    hyperparameters: dict[str | int, typing.Any] = dataclasses.field(default_factory=dict)

    def __getitem__(self, key):
        return self.hyperparameters[key]

    def __setitem__(self, key, value):
        self.hyperparameters[key] = value
        if key not in self.hyperparameter_definitions:
            self.hyperparameter_definitions[key] = HyperparameterDefinition(key, type(value), ...)

    @staticmethod
    def from_function(f: typing.Callable) -> "HyperparameterBuilder":
        escaped_name = string.capwords(f.__name__, sep="_").replace("_", "")
        class_name = f"{escaped_name}Hyperparameters"
        return HyperparameterBuilder(f.__module__, class_name)

    @staticmethod
    def from_model(model: BaseModel) -> "HyperparameterBuilder":
        hyperparameter_definitions: dict[str, HyperparameterDefinition] = {
            name: HyperparameterDefinition(name, field.type_, field.field_info)
            for name, field in model.__fields__.items()
        }
        return HyperparameterBuilder(
            model.__module__, model.__class__.__name__, hyperparameter_definitions, dict(model)
        )

    def get_hyperparameter(self, name: str | int, default: T) -> T:
        return self.hyperparameters.get(name, default)

    def get_type(self, name):
        definition = self.hyperparameter_definitions[name]
        if definition.explicit_type is Undefined:
            return type(self.hyperparameters.get(name, definition.field_info.default))
        else:
            return definition.explicit_type

    def build(self):
        field_definitions = {
            name: (self.get_type(name), definition.field_info)
            for name, definition in self.hyperparameter_definitions.items()
        }

        model = create_model(self.qualname, __module__=self.module, **field_definitions)
        return model(**self.hyperparameters)

    def context(self):
        return HyperparameterContext(self)


@dataclass
class HyperparameterContext:
    hyperparameter_builder: HyperparameterBuilder
    unique_id: int = 0

    @contextmanager
    def scope(self):
        global _hyperparameter_context
        old_hyperparameter_context = _hyperparameter_context
        _hyperparameter_context = self
        try:
            yield
        finally:
            _hyperparameter_context = old_hyperparameter_context


# TODO: replace with a context var
_hyperparameter_context: HyperparameterContext | None = None


class Hyperparameters(BaseModel):
    def __getitem__(self, f: typing.Callable):
        module_escaped_name = HyperparametersBuilder.convert_module_name(f.__module__)
        f_escaped_name = HyperparametersBuilder.convert_function_name(f.__qualname__)

        module = getattr(self, module_escaped_name, None)
        if module is None:
            raise KeyError(f"Module {module_escaped_name} for {f} not found in hyperparameters!")
        f_hyperparameters = getattr(module, f_escaped_name, None)
        if f_hyperparameters is None:
            raise KeyError(f"Function {f_escaped_name} for {f} not found in hyperparameters!")
        return f_hyperparameters

    @staticmethod
    def merge(hyperparameters_iterable: typing.Iterable["Hyperparameters"]) -> "Hyperparameters":
        hyperparameters_builder = HyperparametersBuilder()
        for hyperparameters in hyperparameters_iterable:
            hyperparameters_builder.update(hyperparameters)

        hyperparameters = hyperparameters_builder.build()
        return hyperparameters


@dataclass
class HyperparametersBuilder:
    hyperparameter_builders: defaultdict[str, dict[str, HyperparameterBuilder]] = dataclasses.field(
        default_factory=lambda: defaultdict(dict)
    )

    @staticmethod
    def from_optional_model(hyperparameters: BaseModel | None):
        builder = HyperparametersBuilder()
        if hyperparameters is not None:
            builder.update(hyperparameters)
        return builder

    def __getitem__(self, f: typing.Callable) -> HyperparameterBuilder:
        """
        Look up f in a model created with `create_model`.
        """
        module_escaped_name = HyperparametersBuilder.convert_module_name(f.__module__)
        f_escaped_name = HyperparametersBuilder.convert_function_name(f.__qualname__)

        builder = self.hyperparameter_builders[module_escaped_name].setdefault(
            f_escaped_name, HyperparameterBuilder.from_function(f)
        )
        return builder

    def update(self, hyperparameters: BaseModel):
        """
        Update the hyperparameters from a model created with `create_model`.
        """
        for module_name, module_hparams in hyperparameters:
            for f_name, f_hparams in module_hparams:
                self.hyperparameter_builders[module_name][f_name] = HyperparameterBuilder.from_model(f_hparams)

    @staticmethod
    def convert_module_name(module_name: str):
        # replace . with _ and remove prefix "__"
        module_escaped_name = module_name.replace(".", "_")
        if module_escaped_name.startswith("__"):
            module_escaped_name = module_escaped_name[2:]
        return module_escaped_name

    @staticmethod
    def convert_function_name(f_qual_name: str):
        # convert f_qual_name to a valid class name (. and <. are not allowed)
        f_escaped_name = f_qual_name.replace(".", "_").replace("<", "_").replace(">", "_")
        return f_escaped_name

    def build(self):
        # turn function name into a class name
        class_name = "Hyperparameters"

        fields = {}

        # create a model for each model (first nesting level)
        for module_name, module_hparams in self.hyperparameter_builders.items():
            module_fields = {}
            for f_qual_name, f_builder in module_hparams.items():
                model = f_builder.build()
                module_fields[f_qual_name] = model

            module_class_name = f"{module_name}Module"

            module_model = create_model(module_class_name, __module__=__name__, **module_fields)

            fields[module_name] = module_model()

        hyperparameter_model = create_model(class_name, __module__=__name__, __base__=Hyperparameters, **fields)
        return hyperparameter_model()

    @contextmanager
    def scope(self):
        """
        Context manager that allows to track hyperparameters.
        """
        global _hyperparameters_builder
        old_hyperparameters_builder = _hyperparameters_builder
        _hyperparameters_builder = self

        try:
            yield
        finally:
            _hyperparameters_builder = old_hyperparameters_builder


_hyperparameters_builder: HyperparametersBuilder | None = None


@dataclass
class TrackedFunction(CallableWrapper, typing.Callable[P, T], typing.Generic[P, T]):  # type: ignore
    """
    A callable that can be called with a chat model.
    """

    __wrapped__: typing.Callable[P, T]

    @staticmethod
    def from_function(f: typing.Callable[P, T]):
        tracked_function: TrackedFunction = functools.wraps(f)(
            TrackedFunction(
                f,
            )
        )

        return tracked_function

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if _hyperparameters_builder is None:
            return self.__wrapped__(*args, **kwargs)

        with _hyperparameters_builder[self].context().scope():
            return self.__wrapped__(*args, **kwargs)


def track_hyperparameters(f: typing.Callable[P, T]) -> typing.Callable[P, T]:
    return TrackedFunction.from_function(f)


@dataclass
class HyperparametersScope:
    hyperparameters: BaseModel | None = None

    @contextmanager
    def __call__(self):
        builder = HyperparametersBuilder.from_optional_model(self.hyperparameters)
        with builder.scope():
            yield self
        self.hyperparameters = builder.build()


@contextmanager
def hyperparameters_scope(hyperparameters: BaseModel | None = None) -> typing.Iterator[HyperparametersScope]:
    scope = HyperparametersScope(hyperparameters)
    with scope():
        yield scope



import pytest
from langchain.schema import AIMessage, HumanMessage

from llm_hyperparameters.track_execution import TrackedChatModel
from llm_strategy.chat_chain import ChatChain
from llm_strategy.testing.fake_chat_model import FakeChatModel


def test_chat_chain():
    # Only test ChatChain
    chat_model = FakeChatModel.from_messages(
        [
            [
                HumanMessage(content="Hello"),
                AIMessage(content="World"),
                HumanMessage(content="How are you?"),
                AIMessage(content="Good. How are you?"),
            ],
            [
                HumanMessage(content="Hello"),
                AIMessage(content="World"),
                HumanMessage(content="What's up?"),
                AIMessage(content="Nothing. You?"),
            ],
        ]
    )

    tracked_chat_model = TrackedChatModel(chat_model=chat_model)

    chat_chain = ChatChain(tracked_chat_model, [HumanMessage(content="Hello")])

    assert chat_chain.response == "Hello"

    assert chat_chain.messages == [
        HumanMessage(content="Hello"),
    ]

    chat_chain_2 = ChatChain(tracked_chat_model, [])

    with pytest.raises(AssertionError):
        chat_chain_2.response

    assert chat_chain_2.messages == []

    response, chat_chain_3 = chat_chain_2.query("Hello")

    assert response == "World"

    assert chat_chain_3.messages == [
        HumanMessage(content="Hello"),
        AIMessage(content="World"),
    ]

    assert chat_chain_3.response == "World"
    assert tracked_chat_model.tracked_chats.build_compact_dict() == {
        "children": [],
        "messages": [{"content": "Hello", "role": "user"}, {"content": "World", "role": "assistant"}],
    }

    chat_chain_4 = chat_chain_3.branch()

    assert chat_chain_4.messages == [
        HumanMessage(content="Hello"),
        AIMessage(content="World"),
    ]

    response, chat_chain_5 = chat_chain_4.query("How are you?")
    assert response == "Good. How are you?"
    assert chat_chain_5.messages == [
        HumanMessage(content="Hello"),
        AIMessage(content="World"),
        HumanMessage(content="How are you?"),
        AIMessage(content="Good. How are you?"),
    ]

    assert tracked_chat_model.tracked_chats.build_compact_dict() == {
        "children": [],
        "messages": [
            {"content": "Hello", "role": "user"},
            {"content": "World", "role": "assistant"},
            {"content": "How are you?", "role": "user"},
            {"content": "Good. How are you?", "role": "assistant"},
        ],
    }

    response, chat_chain_6 = chat_chain_4.query("What's up?")
    assert response == "Nothing. You?"
    assert chat_chain_6.messages == [
        HumanMessage(content="Hello"),
        AIMessage(content="World"),
        HumanMessage(content="What's up?"),
        AIMessage(content="Nothing. You?"),
    ]

    assert tracked_chat_model.tracked_chats.build_compact_dict() == {
        "messages": [{"content": "Hello", "role": "user"}, {"content": "World", "role": "assistant"}],
        "children": [
            {
                "children": [],
                "messages": [
                    {"content": "How are you?", "role": "user"},
                    {"content": "Good. How are you?", "role": "assistant"},
                ],
            },
            {
                "children": [],
                "messages": [
                    {"content": "What's up?", "role": "user"},
                    {"content": "Nothing. You?", "role": "assistant"},
                ],
            },
        ],
    }



#  Blackboard-PAGI - LLM Proto-AGI using the Blackboard Pattern
#  Copyright (c) 2023. Andreas Kirsch
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Collection, Iterable, List, Mapping, Optional

from langchain.chat_models.base import BaseChatModel
from langchain.schema import (
    AIMessage,
    BaseMessage,
    ChatResult,
    messages_from_dict,
    messages_to_dict,
)
from pydantic import BaseModel, Field


def dict_to_tuple(d: Mapping[Any, Any]) -> tuple[tuple[Any, Any], ...]:
    """Convert a dict to a tuple of tuples, sorted by key."""
    # Convert values that are dicts to tuples as well.
    return tuple(sorted((k, dict_to_tuple(v) if isinstance(v, dict) else v) for k, v in d.items()))


def tuple_to_dict(t: Iterable[tuple[Any, Any]]) -> dict[Any, Any]:
    """Convert a tuple of tuples to a dict."""
    return {k: tuple_to_dict(v) if isinstance(v, tuple) else v for k, v in t}


def is_prefix_list(prefix_candidate: Collection, messages: Collection) -> bool:
    """Return whether `prefix_candidate` is a prefix of `messages`."""
    if len(prefix_candidate) > len(messages):
        return False
    for prefix_message, message in zip(prefix_candidate, messages):
        if prefix_message != message:
            return False
    return True


class FakeChatModel(BaseChatModel, BaseModel):
    """Fake ChatModel wrapper for testing purposes.

    We can use this to test the behavior of the LLM wrapper without having to actually call the LLM.

    We support an `external_llm` argument, which is an LLM that will be called if the query is not found in the `texts`
    dict. We store the responses. On exit, we deduplicate them and print them to stdout so that they can be copied into
    constructor call for the next run by hand if needed.

    We support stop words, which are words that are removed from the response if they are found. To do so, we store
    the full response (as it is build over time) and return the part before the query and the stop word.

    This also means that there is no non-determinism in the output, which is good for testing, but bad for variance.
    Especially if we want to test the behavior of the LLM wrapper when the LLM is not deterministic. (Create different
    outputs for different calls, for example.)
    """

    messages_tuples_bag: set[tuple] = Field(default_factory=set)
    """The texts to return on call."""
    external_chat_model: BaseChatModel | None = None
    """An external LLM to use if the query is not found."""

    @property
    def _llm_type(self) -> str:
        return "fake"

    @staticmethod
    def from_messages(messages_bag: Collection[list[BaseMessage]]) -> "FakeChatModel":
        messages_tuples_bag = {tuple(dict_to_tuple(m) for m in messages_to_dict(messages)) for messages in messages_bag}
        return FakeChatModel(messages_tuples_bag=messages_tuples_bag)

    async def _agenerate(self, messages: list[BaseMessage], stop: list[str] | None = None, **kwargs) -> ChatResult:
        raise NotImplementedError

    def _generate(self, messages: List[BaseMessage], stop: Optional[List[str]] = None, **kwargs) -> ChatResult:
        raise NotImplementedError

    def __del__(self) -> None:
        # If we have an external LLM, we write out all our responses to stdout so that they can be copied into the
        # constructor call for the next run by hand if needed.
        if self.external_chat_model is not None:
            # Deduplicate the messages (any shared prefixes can be removed)
            self.messages_tuples_bag = {
                messages
                for messages in self.messages_tuples_bag
                if not any(is_prefix_list(messages, other) for other in self.messages_tuples_bag if other != messages)
            }
            print(f"messages_bag = {self.messages_tuples_bag!r}")

    def invoke(self, messages: list[BaseMessage], stop: list[str] | None = None, **kwargs) -> BaseMessage:
        """Return the query if it exists, else print the code to update the query."""
        assert stop is None, "Stop words are not supported for FakeChatModel."

        messages_tuple = tuple(dict_to_tuple(m) for m in messages_to_dict(messages))

        for cached_messages in self.messages_tuples_bag:
            if is_prefix_list(messages_tuple, cached_messages):
                # check that the next message is an AIMessage
                if len(cached_messages) == len(messages_tuple):
                    raise ValueError("No response found in messages_bag.")
                next_message = messages_from_dict([tuple_to_dict(cached_messages[len(messages)])])[0]
                if not isinstance(next_message, AIMessage):
                    raise ValueError("No response found in messages_bag.")
                return next_message

        if self.external_chat_model is not None:
            message = self.external_chat_model.invoke(messages, stop=stop, **kwargs)
            message_tuple = dict_to_tuple(messages_to_dict([message])[0])
            self.messages_tuples_bag.add(tuple(list(messages_tuple) + [message_tuple]))
            return message

        # If no queries are provided, print the code to update the query
        code_snippet = f"""# Add the following to the queries dict:
{messages!r}, # TODO: Append the correct response here
"""
        print(code_snippet)
        raise NotImplementedError("No query provided. Add the following to the queries dict:\n\n" + code_snippet)

    __call__ = invoke



import typing
from typing import Any, Dict, List, Optional

from langchain.schema import BaseMessage, ChatMessage, ChatResult, LLMResult
from langchain_core.language_models import BaseChatModel, BaseLanguageModel, BaseLLM
from llmtracer import TraceNodeKind, trace_calls, trace_object_converter
from llmtracer.object_converter import ObjectConverter
from pydantic import BaseModel, Field

from llm_strategy.chat_chain import ChatChain

T = typing.TypeVar("T")
P = typing.ParamSpec("P")

LangchainInterface: typing.TypeAlias = BaseLanguageModel | ChatChain
LI = typing.TypeVar("LI", BaseLanguageModel, ChatChain)


class ChatTree(BaseModel):
    message: BaseMessage | None
    children: list["ChatTree"]

    @staticmethod
    def create_root():
        return ChatTree(message=None, children=[])

    def insert(self, messages: list[BaseMessage]) -> "ChatTree":
        node = self
        for message in messages:
            for child in node.children:
                if child.message == message:
                    node = child
                    break
            else:
                new_node = ChatTree(message=message, children=[])
                node.children.append(new_node)
                node = new_node
        return node

    def build_compact_dict(self) -> dict:
        """
        Returns a compact JSON representation of the chat tree.
        If we only have one child, we concatenate the messages until we hit a child with more than one child.

        We include the role of the message in the JSON.
        """
        node = self
        messages_json = []
        while True:
            message = node.message
            if message is not None:
                if message.type == "human":
                    role = "user"
                elif message.type == "ai":
                    role = "assistant"
                elif message.type == "system":
                    role = "system"
                elif message.type == "chat":
                    assert isinstance(message, ChatMessage)
                    role = message.role
                else:
                    raise ValueError(f"Unknown message type {message.type}")
                messages_json.append({"role": role, "content": message.content})
            if len(node.children) == 1:
                node = node.children[0]
            else:
                break

        return {
            "messages": messages_json,
            "children": [child.build_compact_dict() for child in node.children],
        }


class PromptTree(BaseModel):
    fragment: str = ""
    children: list["PromptTree"]

    @staticmethod
    def create_root():
        return PromptTree(message=None, children=[])

    def insert(self, text: str) -> "PromptTree":
        node = self
        while len(text):
            for child in node.children:
                if text.startswith(child.fragment):
                    node = child
                    text = text.removeprefix(child.fragment)
                    break
            else:
                new_node = PromptTree(fragment=text, children=[])
                node.children.append(new_node)
                return new_node
        return node

    def build_compact_dict(self) -> dict:
        """
        Returns a compact JSON representation of the chat chain.
        If we only have one child, we concatenate the messages until we hit a child with more than one child.

        We include the role of the message in the JSON.
        """
        node = self
        fragments_json = []
        while True:
            if len(node.fragment):
                fragments_json.append(node.fragment)
            if len(node.children) == 1:
                node = node.children[0]
            else:
                break

        return {
            "fragments": fragments_json,
            "children": [child.build_compact_dict() for child in node.children],
        }


class TrackedLLM(BaseLLM):
    llm: BaseLLM
    tracked_prompts: PromptTree = Field(default_factory=PromptTree.create_root)

    @trace_calls(name="TrackedLLM", kind=TraceNodeKind.LLM, capture_args=True, capture_return=True)
    def invoke(self, prompt: str, stop: Optional[List[str]] = None, **kwargs) -> str:
        node = self.tracked_prompts.insert(prompt)
        response = self.llm.invoke(prompt, stop, **kwargs)
        node.insert(response)
        return response

    __call__ = invoke

    def _generate(self, prompts: List[str], stop: Optional[List[str]] = None, **kwargs) -> LLMResult:
        raise NotImplementedError()

    async def _agenerate(self, prompts: List[str], stop: Optional[List[str]] = None, **kwargs) -> LLMResult:
        raise NotImplementedError()

    @property
    def _llm_type(self) -> str:
        return self.llm._llm_type


class TrackedChatModel(BaseChatModel):
    chat_model: BaseChatModel
    tracked_chats: ChatTree = Field(default_factory=ChatTree.create_root)

    @property
    def _llm_type(self) -> str:
        return self.chat_model._llm_type

    def dict(self, **kwargs: Any) -> Dict:
        return self.chat_model.dict(**kwargs)

    @trace_calls(name="TrackedChatModel", kind=TraceNodeKind.LLM, capture_args=True, capture_return=True)
    def invoke(self, messages: List[BaseMessage], stop: Optional[List[str]] = None, **kwargs) -> BaseMessage:
        response_message = self.chat_model.invoke(messages, stop, **kwargs)
        self.tracked_chats.insert(messages + [response_message])
        return response_message

    __call__ = invoke

    def _generate(self, messages: List[BaseMessage], stop: Optional[List[str]] = None, **kwargs) -> ChatResult:
        chat_result = self.chat_model._generate(messages, stop, **kwargs)
        node = self.tracked_chats.insert(messages)
        for generation in chat_result.generations:
            node.insert([generation.message])
        return chat_result

    async def _agenerate(self, messages: List[BaseMessage], stop: Optional[List[str]] = None, **kwargs) -> ChatResult:
        chat_result = await self.chat_model._agenerate(messages, stop, **kwargs)
        node = self.tracked_chats.insert(messages)
        for generation in chat_result.generations:
            node.insert([generation.message])
        return chat_result


def track_langchain(language_model_or_chat_chain: LI) -> LI:
    if isinstance(language_model_or_chat_chain, ChatChain):
        return ChatChain(
            chat_model=TrackedChatModel(chat_model=language_model_or_chat_chain.chat_model),
            messages=language_model_or_chat_chain.messages,
        )
    elif isinstance(language_model_or_chat_chain, BaseLLM):
        return TrackedLLM(llm=language_model_or_chat_chain)
    elif isinstance(language_model_or_chat_chain, BaseChatModel):
        return TrackedChatModel(language_model=language_model_or_chat_chain)
    else:
        raise ValueError(f"Unknown language model type {type(language_model_or_chat_chain)}")


def get_tracked_chats(chat_model_or_chat_chain: ChatChain | TrackedChatModel) -> dict:
    if isinstance(chat_model_or_chat_chain, ChatChain):
        model = chat_model_or_chat_chain.chat_model
    elif isinstance(chat_model_or_chat_chain, TrackedChatModel):
        model = chat_model_or_chat_chain
    else:
        raise ValueError(f"Unknown language model type {type(chat_model_or_chat_chain)}")
    assert isinstance(model, TrackedChatModel)
    return model.tracked_chats.build_compact_dict()


@trace_object_converter.register_converter()
def _convert_llm(llm: BaseLLM, converter: ObjectConverter) -> dict:
    return converter(llm.dict(), converter)  # type: ignore


@trace_object_converter.register_converter()
def _convert_chat_model(chat_model: BaseChatModel, converter: ObjectConverter) -> dict:
    return dict(_type=chat_model.__class__.__name__)


@trace_object_converter.register_converter()
def _convert_tracked_chat_model(chat_model: TrackedChatModel, converter: ObjectConverter) -> dict:
    return dict(_type=chat_model.chat_model.__class__.__name__)



import dataclasses
import typing
from dataclasses import dataclass
from typing import Tuple, cast

from langchain.chat_models.base import BaseChatModel
from langchain.output_parsers import PydanticOutputParser
from langchain.schema import BaseMessage, HumanMessage
from pydantic import BaseModel, create_model

T = typing.TypeVar("T")
B = typing.TypeVar("B", bound=BaseModel)


@dataclass
class ChatChain:
    chat_model: BaseChatModel
    messages: list[BaseMessage]

    @property
    def response(self) -> str:
        assert len(self.messages) >= 1
        return cast(str, self.messages[-1].content)

    def append(self, messages: list[BaseMessage]) -> "ChatChain":
        return dataclasses.replace(self, messages=self.messages + messages)

    def __add__(self, other: list[BaseMessage]) -> "ChatChain":
        return self.append(other)

    def query(self, question: str, model_args: dict | None = None) -> Tuple[str, "ChatChain"]:
        """Asks a question and returns the result in a single block."""
        # Build messages:
        messages = self.messages + [HumanMessage(content=question)]
        model_args = model_args or {}
        reply = self.chat_model.invoke(messages, **model_args)
        messages.append(reply)
        return cast(str, reply.content), dataclasses.replace(self, messages=messages)

    def enforce_json_response(self, model_args: dict | None = None) -> dict:
        model_args = model_args or {}
        # Check if the language model is of type "openai" and extend model args with a response format in that case
        model_dict = self.chat_model.dict()
        if "openai" in model_dict["_type"] and model_dict.get("model_name") in (
            "gpt-4-1106-preview",
            "gpt-3.5-turbo-1106",
        ):
            model_args = {**model_args, "response_format": dict(type="json_object")}
        return model_args

    def structured_query(
        self, question: str, return_type: type[B], model_args: dict | None = None
    ) -> Tuple[B, "ChatChain"]:
        """Asks a question and returns the result in a single block."""
        # Build messages:

        if typing.get_origin(return_type) is typing.Annotated:
            return_info = typing.get_args(return_type)
        else:
            return_info = (return_type, ...)

        output_model = create_model("StructuredOutput", result=return_info)
        parser: PydanticOutputParser = PydanticOutputParser(pydantic_object=output_model)
        question_and_formatting = question + "\n\n" + parser.get_format_instructions()

        reply_content, chain = self.query(question_and_formatting, self.enforce_json_response(model_args))
        parsed_reply: B = typing.cast(B, parser.parse(reply_content))

        return parsed_reply, chain

    def branch(self) -> "ChatChain":
        return dataclasses.replace(self, messages=self.messages.copy())



from typing import Any, Dict, List, Optional

from langchain.chat_models.base import BaseChatModel
from langchain.llms import BaseLLM
from langchain.schema import AIMessage, BaseMessage, ChatMessage, ChatResult, LLMResult
from typing_extensions import override


class ChatModelAsLLM(BaseLLM):
    chat_model: BaseChatModel

    @override
    def dict(self, **kwargs: Any) -> Dict:
        return self.chat_model.dict()

    @override
    def invoke(self, prompt: str, *, stop: Optional[List[str]] = None, **kwargs) -> str:
        response = self.chat_model.call_as_llm(prompt, stop=stop, **kwargs)
        return response

    __call__ = invoke

    @override
    def _generate(self, prompts: List[str], *, stop: Optional[List[str]] = None, **kwargs) -> LLMResult:
        raise NotImplementedError()

    @override
    async def _agenerate(self, prompts: List[str], *, stop: Optional[List[str]] = None, **kwargs) -> LLMResult:
        raise NotImplementedError()

    @override
    @property
    def _llm_type(self) -> str:
        return self.chat_model._llm_type


class LLMAsChatModel(BaseChatModel):
    llm: BaseLLM

    @override
    def dict(self, **kwargs: Any) -> Dict:
        return self.llm.dict()

    @staticmethod
    def convert_messages_to_prompt(messages: list[BaseMessage]) -> str:
        prompt = ""
        for message in messages:
            if message.type == "human":
                role = "user"
            elif message.type == "ai":
                role = "assistant"
            elif message.type == "system":
                role = "system"
            elif message.type == "chat":
                assert isinstance(message, ChatMessage)
                role = message.role.capitalize()
            else:
                raise ValueError(f"Unknown message type {message.type}")
            prompt += f"<|im_start|>{role}\n{message.content}<|im_end|>"
        prompt += "<|im_start|>assistant\n"
        return prompt

    @override
    @property
    def _llm_type(self) -> str:
        return self.llm._llm_type

    @override
    def invoke(self, messages: List[BaseMessage], *, stop: Optional[List[str]] = None, **kwargs) -> BaseMessage:
        prompt = self.convert_messages_to_prompt(messages)
        stop = [] if stop is None else list(stop)
        response = self.llm.invoke(prompt, stop=["<|im_end|>"] + stop, **kwargs)
        return AIMessage(content=response)

    __call__ = invoke

    @override
    def _generate(self, messages: List[BaseMessage], *, stop: Optional[List[str]] = None, **kwargs) -> ChatResult:
        raise NotImplementedError()

    @override
    async def _agenerate(
        self, messages: List[BaseMessage], *, stop: Optional[List[str]] = None, **kwargs
    ) -> ChatResult:
        raise NotImplementedError()



import types
import typing
from dataclasses import dataclass

P = typing.ParamSpec("P")
T = typing.TypeVar("T")


@dataclass
class CallableWrapper:
    """
    A functor that wraps a callable and forwards all attributes to the wrapped callable.
    """

    __wrapped__: object

    def __get__(self, instance: object, owner: type | None = None) -> typing.Callable:
        """Support instance methods."""
        if instance is None:
            return self

        # Bind self to instance as MethodType
        return types.MethodType(self, instance)

    def __getattr__(self, item):
        return getattr(self.__wrapped__, item)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def __dir__(self) -> typing.Iterable[str]:
        # merge the attributes of the wrapped object with the attributes of the wrapper
        self_dir = list(super().__dir__())
        wrapped_dir = list(dir(self.__wrapped__))
        return list(set(self_dir + wrapped_dir))
